/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

#include <tyros.h>
#include <ntx.h>

void less_file(char *filename);

int main(int argc, char *argv[])
{
  if(argc == 1) {
    printf("less: nothing to display");
  } else {
    less_file(argv[1]);
  }
}


#define CON_SIZE (CON_WIDTH*CON_HEIGHT)
#define MAX_FILE_SIZE 0x10000



void less_file(char *filename)
{
  int fd;
  ntx_stat_t stbuf;
  char key[0x10];
  int cper;
  int nlines;
  int total_lines;
  int cline;
  int ltop;

  fd = open(filename, 0);

  if(fd == -1) {
    printf("Can't open %s!\n", filename);
    return;
  }

  fstat(fd, &stbuf);

  cline = 0;

  for(;;) {
    total_lines = show_at_position(fd, stbuf.st_size, cline);
    
    if((cline + CON_HEIGHT -1) > total_lines) {
      ltop = total_lines - cline;
    } else {
      ltop = CON_HEIGHT - 1;
    }
    
    printf("-- Viewing %s (line %i+%i of %i) >", filename, cline, ltop, total_lines);
    
    fgets(key, 10, stdin);
    
    switch(key[0]) {
    case 'q':
      return;
      break;
      
    default:
      cline += (CON_HEIGHT - 1);
      if(cline > total_lines) {
	cline = total_lines;
      }
      break;
    }
  }
}

int show_at_position (int fd, int size, int tline)
{
  char buffer[MAX_FILE_SIZE];
  int num;
  int start;
  int nlines;
  int bsize;
  int bbegin;


  num = read (fd, buffer, MAX_FILE_SIZE);
  
  nlines = count_lines(buffer);

  
  bbegin = find_line(buffer, tline);

  bsize = get_block_size (buffer, tline, tline + CON_HEIGHT - 1);

  //  printf("LLChar: %i\n", find_line(buffer, tline + CON_WIDTH));
  
  //  printf("BlockSize: %i\n", bsize);
  
  write(stdout, &buffer[bbegin], bsize);
  
  return nlines;
}


int count_lines (char *buffer)
{
  int nlines;

  nlines = 0;
  
  while(*buffer != 0) {
    
    if(*buffer == '\n') {
      nlines++;
    }

    buffer++;
  }

  return nlines;
}

int find_line (char *buffer, int line)
{
  int nlines;
  int i;
  int last;

  nlines = 0;
  i = 0;
  last = 0;

  while(*buffer != 0) {
    
    if(nlines == line) {
      return i;
    }
    
    if(*buffer == '\n' || ((i - last) == (CON_WIDTH))) {
      nlines++;
      last = i;
    }
    
    buffer++;
    i++;
  }


  return i;
}

int get_block_size (char *buffer, int first_line, int last_line)
{
  return (find_line(buffer, last_line) - find_line(buffer, first_line));
}
