/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

#include <stdio.h>
#include "elf.h"

int main(int argc, char *argv[])
{
  int fd;
  char *buffer;
  int size;
  
  
  fd = open("libshare.so", 0);
  
  if(fd == -1) {
    printf("error opening exe.elf!\n");
    return 0;
  }
  
  buffer = malloc(0x10000);
  
  size = read(fd, buffer, 0x10000);
  
  printf("read %i bytes\n", size);

  load_elf(buffer);

  close(fd);
}

char SYMTYPENAME[][20] = 
  {
    "NOTYPE ",
    "OBJECT ",
    "FUNC   ",
    "SECTION",
    "FILE   "
  };

char SYMBINDNAME[][20] =
  {
    "LOCAL ",
    "GLOBAL",
    "WEAK"
  };


int load_elf (void *image)
{
  int i, j;
  void * ptask_entry;
  unsigned int back, back2, s;
  unsigned int pagedir;
  unsigned int total_pages;
  unsigned int total_size;
  unsigned int phys_page;
  unsigned int psize;
  struct elf32_hdr *EH = (struct elf32_hdr *)image;
  struct elf32_phdr *EPH;
  struct elf32_shdr *ESH, *shstrtab, *dynstr, *dynsect, *dynsym;
  struct elf32_shdr *dynstrtab, *symtab, *rel, *rela, *relsymtab;
  struct elf32_sym *dsymtable, *relsymtable;
  struct elf32_rel *reltable;
  int dsym_count, relsym_count, rel_count;
  char *shstrings, *dsymstr;
  struct dynamic *dyndyn;
  int dyn_count;
  char *name;
  char *strings;

  printf("loading elf image....\n");
  
  EPH = (struct elf32_phdr *)(image + EH->e_phoff);
  ESH = (struct elf32_shdr *)(image + EH->e_shoff);

  printf("Program header table is found at %X\n", (unsigned int) EPH - (unsigned int)image);
  printf("Sections header table is found at %X\n", (unsigned int) ESH - (unsigned int)image);
  
  shstrtab = &ESH[EH->e_shstrndx];

  shstrings = image + shstrtab->sh_offset;
  
  shstrtab = &ESH[EH->e_shstrndx];
  
  for(i = 0; i < EH->e_shnum; i++) {
    
    name = shstrings + ESH[i].sh_name;

    
    if(ESH[i].sh_type == SHT_DYNSYM) {
      printf(".dynsym found\n");
      dynsym = &ESH[i];
    }
    
    if(ESH[i].sh_type == SHT_REL) {
      printf(".rel found\n");
      rel = &ESH[i];
      
      relsymtab = &ESH[rel->sh_link];
      
      relsymtable = image + relsymtab->sh_offset;
      
      relsym_count = relsymtab->sh_size/sizeof(struct elf32_sym);
      rel_count = rel->sh_size/sizeof(struct elf32_rel);
      
      printf("relocation applies to (%i) '%s' (%i symbols, %i relocations)\n", 
	     rel->sh_info,
	     ESH[rel->sh_info].sh_name + shstrings,
	     relsym_count,
	     rel_count);
      reltable = image + rel->sh_offset;
      
      strings = image + ESH[relsymtab->sh_link].sh_offset;
      
      
      
      for(j = 0; j < rel_count; j++) {
	printf("relocation: [type: %i], sym: '%s', off: %.8X\n",
	       ELF32_R_TYPE(reltable[j].r_info),
	       strings + relsymtable[ELF32_R_SYM(reltable[j].r_info)].st_name,
	       reltable[j].r_offset);
      }
      
    }

    if(ESH[i].sh_type == SHT_RELA) {
      printf(".rela found\n");
      rela = &ESH[i];
      
      
    }
  
    if(ESH[i].sh_type == SHT_SYMTAB) {
      printf(".symtab found\n");
      symtab = &ESH[i];
    }
    
    if(ESH[i].sh_type == SHT_DYNAMIC) {
      printf(".dynamic found\n");
      dynsect = &ESH[i];
    }
    
  }
  
  dynstr = &ESH[dynsect->sh_link];

  dyndyn = image + dynsect->sh_offset;

  dyn_count = dynsect->sh_size/sizeof(struct dynamic);
  
  dsymtable = image + dynsym->sh_offset;
  
  dsymstr = image + dynstr->sh_offset;
  
  dsym_count = dynsym->sh_size/sizeof(struct elf32_sym);
  
  printf("file contains %i dynamic symbols\n", dsym_count);
  
  for(i = 0; i < dsym_count; i++) {
    printf("symbol: [type: %s | %s off: %.8X size %.8X] %s\n", 
	   SYMTYPENAME[ELF32_ST_TYPE(dsymtable[i].st_info)],
	   SYMBINDNAME[ELF32_ST_BIND(dsymtable[i].st_info)],
	   dsymtable[i].st_value,
	   dsymtable[i].st_size,
	   dsymstr + dsymtable[i].st_name);
  }
  
  printf("dynamic table contains %i entries\n", dyn_count);
  
  
  
  for(i = 0; i < dyn_count; i++) {
    
    switch(dyndyn[i].d_tag) {
    case DT_NEEDED:
      printf("required library: %s\n", dsymstr + dyndyn[i].d_un.d_val);
      break;
      
    case DT_STRTAB:
      //      printf("DT_STRTAB: %X\n", dyndyn[i].d_un.d_val);
      break;
    }
    
    /*    if(dyndyn[i].d_tag == DT_NEEDED) {
      printf("DT_NEEDED: %X\n", dyndyn[i].d_un.d_val);
      //      printf("%s", dyndyn[i].d_un.d_val);
      }*/
  }
  
}

int deal_dynamic (struct dynamic *data, int size, char *strtab)
{
  int entries = size/sizeof(struct dynamic);
  int i;
  

  printf("dynamic section consists %i bytes or %i entries\n", size, entries);
  
  for(i = 0; i < entries; i++) {
    if(data[i].d_tag == DT_NEEDED) {
      printf("needed libs:\n");
      
      while(data[i].d_un.d_val != NULL) {
	printf("string: %X\n", data[i].d_un.d_val);
      }
      //printf("entry %i: type: %X\n", i, data[i].d_tag);
    }
  }
}

void dump(unsigned char *addr, int num)
{
  int i;
  printf("%.8X: ", addr);
  for(i=1;i<=num;i++){
    printf("%.2X", *(addr+i-1));   
    putchar(' ');
    if(i%16 == 0){
      putchar('\n');
      
      if(i!=num){
	printf("%.8X: ",(addr+i));
      }    

    }
  }
}

