/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

#include <tyros.h>
#include <ntx.h>
#include <msgs.h>
#include <rpc.h>
#include <ntxbase.h>
#include <core.h>
#include <vfs.h>
#include <blkdev.h>


blkdev_driver_t *blkdevs; //block device drivers table

void init_blkdevs()
{
int i;
blkdevs = alloc_pmem(mem_to_pages(sizeof(blkdev_driver_t)*MAX_BLKDEV));
for(i=0; i<MAX_BLKDEV; i++)
 {
 blkdevs[i].name = NULL;
 blkdevs[i].driver = NULL;
 }
}

int is_blkdev(int major)
{
if(blkdevs[major].driver != NULL)
 return 0;
return -1;
}

int register_blkdev_driver(int major, char* name, void* driver)
{
if(blkdevs[major].driver != NULL)
 return -1;
else
 {
 printf("register_blkdev_driver: major = %d, name = '%s'\n", major, name);
 blkdevs[major].driver = driver;
 blkdevs[major].name = name;
 }
return 0;
}

int unregister_blkdev_driver(int major)
{
if(blkdevs[major].driver != NULL)
 {
 blkdevs[major].driver = NULL;
 blkdevs[major].name = NULL;
 return 0;
 }
else
 return -1;
}

int blkdev_read (int major, int minor, fsize_t start, fsize_t size, void *buffer)
{
blkdev_info_t info;
  
if(blkdev_info(major, minor, &info) == -1) 
 {
 printf("blkdev_read: error querying device information\n");
 return -1;
 } 

printf("blkdev_read: id(%i, %i); start=%i, size=%i\n", 
		major, minor, start, size);
 
int start_block = start / /*info.block_size*/ 512;
int start_block_offset = start % /*info.block_size*/ 512;
int block_count = size / /*info.block_size*/ 512 + 1;

if(block_count > info.block_count)
 return -1;

int (*driver) (int minor, blkdev_operation_t op);
driver = blkdevs[major].driver;

blkdev_operation_t op;
op.cmd = BLKDEV_READ;
op.start_block = start_block;
op.block_count = block_count;
op.buf = alloc_pmem(mem_to_pages(block_count * info.block_size));

int r = driver(minor, op);
if(r == -1)
 return -1;
 
memcpy(op.buf, buffer, size);
return 0;
}

int blkdev_info (int major, int minor, blkdev_info_t *info)
{
blkdev_operation_t op;
int (*driver) (int minor, blkdev_operation_t op);
driver = blkdevs[major].driver;

op.cmd = BLKDEV_INFO;
op.buf = info;

int r = driver(minor, op);
if(r == -1)
 return -1;
return 0;
}