/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

/*

TODO: fix stack memory allocating in load_thread() 

*/


#include <tyros.h>
#include <assert.h>
#include "core.h"
#include "process.h"
#include "ntxbase.h"

pstable_t *pstable;
process_t *me;

//actually this func should never be executed
//(if it is, our scheduler is broken)
void task0()
{
panic("task0 is alive, scheduler failed\n");
}

extern int free_memory;

int load_process (void *image, int size, int pl)
{
  process_t *process;
  int tid;
  
  isem_disable();

  if(check_executable(image)) {
  
    process = process_alloc();
    process_init (process);
    
    tid = process_load (process, image, size, pl);
    pstable_insert (pstable, tid, process);
    
  } else {
   
    tid = -1;
    err("load_process: executable check failed\n");
  }
  
  isem_enable();
  return tid;
}

process_t *load_thread (void *entry)
{
  task_t *task = tasks.task[tasks.current];
  process_t *process;
  int i;
  unsigned int pagedir;
  unsigned int newtid, tid;
  unsigned int stack_start;
  task_image_t task_image;
  
  isem_disable();
    
  tid = tasks.current;

  /* determine current process's pagedir - thread will have the same */
  pagedir = tasks.task[tid]->pagedir;

  task_image.start = entry;
  task_image.size = 0x1000; // ?
  
  newtid = add_task(task_image, pagedir, tasks.task[tid]->pl);
  
  //we need new stack frame for the thread
  
  //BUGGY! Should allocate new virtual address space using heap_alloc_virtual() */
  stack_start = tasks.task[newtid]->stack0->esp - USER_SPACE_FOR_STACK;
  
  tasks.task[newtid]->stack0->esp = stack_start;
  tasks.task[newtid]->stack0->eip = (int) entry;

  for(i = 0; i < USER_SPACE_FOR_STACK; i+=PAGE_SIZE) {
    set_pte(stack_start - i, pagedir, 0x7, VM_PAGE_ALLOCATED);
  }
  
  
  
  process = process_alloc();
  process_init (process);
  
  pstable_insert(pstable, newtid, process);
  
  isem_enable();
  return process;
}

int child_exit_fork();

extern unsigned int syscall_return;
extern unsigned int syscall_ebp;
extern unsigned int syscall_esp;

int process_fork(void)
{
  int new_pagedir, current_pagedir;
  int newtid;
  task_image_t task_image;
  
  new_pagedir = alloc_pmem(1);

  current_pagedir = me->pagedir;

  copy_pagedir (new_pagedir, current_pagedir);
  
  task_image.start = 0;
  task_image.size = 0;

  newtid = add_task(task_image, new_pagedir, tasks.task[tasks.current]->pl);
 
  tasks.task[newtid]->stack0->eip = syscall_return;
  tasks.task[newtid]->stack0->ebp = syscall_ebp;
  tasks.task[newtid]->stack0->esp = syscall_esp;
  tasks.task[newtid]->stack0->eax = 0;
  
  
  return newtid;
}

void __child_exit_fork()
{
  asm("child_exit_fork:");
  asm("xorl %eax, %eax");
  asm("ret");
}

void exit(int code)
{
  kill_task(tasks.current);
  give_up();
}

int waitpid (int tid, int *status, int options)
{
  task_t *task = tasks.task[tasks.current];

  //  printf("WAITPID %i called by task %i\n",tid, tasks.current);
  
  task->state = TS_SUSPENDED;
  task->reason_why = RS_WAITING_PID_DIE;
  task->recipient = tid;
  
  give_up();
  
  assert(task->state != TS_SUSPENDED);

  return tid;
}

int execve (char *filename, char *argv[], char *envp[])
{
  /*  if(resolve_path (filename) == NULL) {
    return -1;
  }
  */
  panic("execve called!");
  for(;;);
/*char *PATH[0x100];
char buffer[PATH_MAX + 0x10];
char *start;
int argc;
int i;
int finish;
char *ret = NULL;;
file_t f;
char *envpath;
char newpath[PATH_MAX + 0x10];
char progname[PATH_MAX + 0x10];
int child;
  
if(*filename != '/') 
 {
 envpath = getenv("PATH");
 
 if(envpath == NULL) 
  {
  printf("PATH = NULL!\n");
  //strcpy(progname, filename);
  //    return 0;
  } 
 else 
  {
    //    strcpy(progname, filename);
    //    return 0;
  }

 strncpy(buffer, envpath, PATH_MAX);

 argc = 0;
 i = 0;
 finish = 0;

 start = buffer;
 for(;;) 
  {
  if(buffer[i] == ';' || buffer[i] == 0) 
   {
   if(buffer[i] != 0) 
    {
    buffer[i] = 0;
    PATH[argc] = start;
    argc++;
    start = &buffer[i + 1];
    } 
   else 
    {
    PATH[argc] = start;
    argc++;
    start = &buffer[i + 1];
    finish = 1;
    }      
   if(*start == 0 || finish == 1) 
    break;
   }
  i++;
  }
 for(i = 0; i < argc; i++) 
  {
  strncpy(newpath, PATH[i], PATH_MAX);
  strncat(newpath, "/", 3);
  strncat(newpath, filename, PATH_MAX);

  child = launch(newpath, argv, envp);

  if(child != -1) 
   return child;
  } 
 }
return launch(filename, argv, envp); */
}

void startup_tasks()
{
  int tid;
  task_image_t task_image;

  process_t *nameserver, *console, *ramdisk,  *idle;

  task_image.start = &task0;
  task_image.size = 1000; //doesn't matter
  
  /* we don't use generic load_process() because
     idle task is executed in kernel address space */
  tid = add_task(task_image, hal.pagedir, 0);
  tasks.task[tid]->stack0->ds = SYS_DATA_SELECTOR;
  idle = process_alloc();
  process_init(idle);
  pstable_insert(pstable, tid, idle);
 
  me = idle;
}


void core_init()
{  
init_core_ints();
pstable = pstable_alloc();

init_blkdevs();
}

extern void int_timer;

void core_entry()
{
char **envp;
char **argv;
char buffer[0x1000];

printf("CORE is in action.\n");

init_mallocator();
init_extensions();

core_init();
init_syscalls();
startup_tasks();
  
init_ntxbase();
init_gui();
  
envp = malloc(8);
argv = malloc(8);
envp[0] = "PATH=/mnt/rd";
envp[1] = NULL;

argv[0] = "/mnt/rd/startup.ntx";
argv[1] = NULL;

launch("/mnt/rd/startup.ntx", argv, envp);
  
i_disable();
i_install (0x20, &int_timer, RING0 | INTERRUPT_GATE);
i_enable();

tasks.task[0]->state = TS_SUSPENDED;

for(;;);
panic("core_entry hasn't died, scheduler failed\n");
}


/*
  does not switches really, but:
  continues execution in context of new task
  (new pagedir)
*/
void switch_task(int to)
{
  tasks.current = to;
  load_cr3(tasks.task[to]->pagedir);
}


/*  point 'me' to process currently executing
    (get called from HAL scheduler) */


void switch_core_task()
{
  int newtid;

  newtid = tasks.current;
  
  me = pstable->process[newtid];

  
  pstable->process[newtid]->pagedir = tasks.task[tasks.current]->pagedir;
}

void kill_core_task (int tid)
{
  process_cleanup (pstable->process[tid]);
}


void *morecore (int npages)
{
  void *allocated;
  
  allocated = heap_alloc_virtual (me->heap, npages);
  init_for_demand (allocated, npages, me->pagedir);
  
  //  printf("Requested to allocate %i pages of memory, allocated: %X\n", npages, allocated);
  
  return allocated;
}

