/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

#include <tyros.h>
#include <ntx.h>

#define DEVCACHE_MAX_ENTRIES 1024

typedef struct device_cache_struct {
  file_t device;
  void *addr;
} device_cache_t;

device_cache_t *devcache;
int devcache_entries;

void init_devcache()
{
int i;

devcache = alloc_pmem(mem_to_pages(sizeof(device_cache_t)*DEVCACHE_MAX_ENTRIES));
for(i = 0; i < DEVCACHE_MAX_ENTRIES; i++)
 {
 devcache[i].device.id[0] = NULL;
 devcache[i].addr = NULL;
 }
}

void *alloc_device_cache (file_t device, int npages)
{
int i;
for(i = 0; i < DEVCACHE_MAX_ENTRIES; i++)
 {
 if(devcache[i].addr == NULL) 
  {
  devcache[i].addr = alloc_pmem(npages);
  devcache[i].device = device;
  return devcache[i].addr;
  }
 } 
}

void free_device_cache (file_t device)
{
int i;
for(i = 0; i < DEVCACHE_MAX_ENTRIES; i++) 
 if(!filecmp(device, devcache[i].device))
  {
  free_pmem(devcache[i].addr);
  devcache[i].addr = NULL;
  return;
  }
}

void *access_device_cache (file_t device)
{
int i;
for(i = 0; i < DEVCACHE_MAX_ENTRIES; i++) 
 {
 if(!filecmp(device, devcache[i].device))
  return devcache[i].addr;
 }
return NULL;
}

int filecmp(file_t s, file_t d)
{
int i;
for(i = 0; i < MOUNTS_MAX; i++) 
 if(s.id[i] != d.id[i]) 
  {
  return 1;
  }
 else
  {
  if(s.id[i] == -1)
   return 0;
  }
return 0;
}
