/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

#include <tyros.h>
#include <ntx.h>
#include <exec.h>
#include "elf.h"

task_image_t load_elf_dynamic (process_t *process, void *image, int size)
{
  int i;
  void * ptask_entry;
  unsigned int pagedir;
  task_image_t task_image;
  unsigned int total_pages;
  unsigned int total_size;
  unsigned int phys_page;
  unsigned int psize;
  struct elf32_hdr *EH = (struct elf32_hdr *)image;
  struct elf32_phdr *EPH;
  struct elf32_shdr *ESH, *strtab;
  void *linker_image;
  fsize_t linker_size;
  char *linker_filename = "/mnt/rd/ntxdl.so";
  file_t linker_file;
  ntx_stat_t statbuf;

  printf("special interpreter is required (hope it is ntxdl.so :))\n");
  
  task_image.start = image;
  task_image.size = size;
  //  task_image.entry = &ntx_dynamic_linker;
  
  
  
  linker_file = resolve_path(linker_filename);

  if(linker_file.id[0] == 0) {
    printf("Neutronix Dynamic Linker not found!\n");
    task_image.size = -1;
    return task_image;
  } 
  
  stat_file (linker_file, &statbuf);
  linker_size = statbuf.st_size;

  linker_image = alloc_pmem(mem_to_pages(linker_size));
  
  read_file (linker_file, 0, linker_size, linker_image);  
  
  task_image = load_elf_image(process, linker_image, linker_size);
  
  free_pmem(linker_image);
  
  

  return task_image;
  //  printf("loading elf image....\n");
 
  pagedir = process->pagedir;
  
  EPH = (struct elf32_phdr *)(image + EH->e_phoff);
  ESH = (struct elf32_shdr *)(image + EH->e_shoff);

  //  printf("Program header table is found at %X\n", (unsigned int) EPH - (unsigned int)image);
  //  printf("Sections header table is found at %X\n", (unsigned int) ESH - (unsigned int)image);

  /* determine if the program needs special interpreter (such as dynamic linker) */
  
  for(i = 0; i < EH->e_phnum; i++) {
    if(EPH[i].p_type == PT_INTERP) {
      for(;;);
    }
  }
  
  total_size = 0;

  while (EH->e_phnum--) {
    if (EPH->p_type == PT_LOAD) {
      //   memcpy (EPH->p_vaddr, bin + EPH->p_offset, EPH->p_filesz);
      //      printf("Loading section (%X, size %i)\n", image + EPH->p_offset, EPH->p_filesz);
      
      psize = mem_to_pages(EPH->p_filesz);
      
      for(i = 0; i < psize; i++) {
	phys_page = alloc_pmem(1);
	memcpyl(phys_page, image + EPH->p_offset + i*PAGE_SIZE, PAGE_SIZE/4); 
	//	printf("Mapping %X to %X\n", EPH->p_vaddr + i*PAGE_SIZE, phys_page);
	//	dump(phys_page, 0x30);
	map_memory_page (EPH->p_vaddr + i*PAGE_SIZE, phys_page, pagedir, 0x7);
      }
      
      total_size += EPH->p_filesz;
    }
    EPH = (struct elf32_phdr *)((unsigned char *)EPH + EH->e_phentsize);
  }
  
  
  task_image.start = image;
  task_image.size = total_size;
  task_image.entry = USER_TASK_ENTRY;
  
  return task_image;
  
}

task_image_t load_elf_image (process_t *process, void *image, int size)
{
  int i;
  void * ptask_entry;
  unsigned int pagedir;
  task_image_t task_image;
  unsigned int total_pages;
  unsigned int total_size;
  unsigned int phys_page;
  unsigned int psize;
  struct elf32_hdr *EH = (struct elf32_hdr *)image;
  struct elf32_phdr *EPH;
  struct elf32_shdr *ESH, *strtab;

  //  printf("loading elf image....\n");
 
  pagedir = process->pagedir;
  
  EPH = (struct elf32_phdr *)(image + EH->e_phoff);
  ESH = (struct elf32_shdr *)(image + EH->e_shoff);

  //  printf("Program header table is found at %X\n", (unsigned int) EPH - (unsigned int)image);
  //  printf("Sections header table is found at %X\n", (unsigned int) ESH - (unsigned int)image);

  /* determine if the program needs special interpreter (such as dynamic linker) */
  
  for(i = 0; i < EH->e_phnum; i++) {
    if(EPH[i].p_type == PT_INTERP) {
      return load_elf_dynamic (process, image, size);
    }
  }

  total_size = 0;

  while (EH->e_phnum--) {
    if (EPH->p_type == PT_LOAD) {
      //   memcpy (EPH->p_vaddr, bin + EPH->p_offset, EPH->p_filesz);
      //      printf("Loading section (%X, size %i)\n", image + EPH->p_offset, EPH->p_filesz);
      
      psize = mem_to_pages(EPH->p_filesz);
      
      for(i = 0; i < psize; i++) {
	phys_page = alloc_pmem(1);
	memcpyl(phys_page, image + EPH->p_offset + i*PAGE_SIZE, PAGE_SIZE/4); 
	//	printf("Mapping %X to %X\n", EPH->p_vaddr + i*PAGE_SIZE, phys_page);
	//	dump(phys_page, 0x30);
	map_memory_page (EPH->p_vaddr + i*PAGE_SIZE, phys_page, pagedir, 0x7);
      }
      
      total_size += EPH->p_filesz;
    }
    EPH = (struct elf32_phdr *)((unsigned char *)EPH + EH->e_phentsize);
  }


  task_image.start = image;
  task_image.size = total_size;
  task_image.entry = USER_TASK_ENTRY;

  return task_image;

}
