/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

#include <tyros.h>
#include <ntx.h>
#include <ntxbase.h>
#include <vfs.h>
#include <filesystems.h>

#define CTLFS_RESTART 1
#define CTLFS_KEX 2
#define CTLFS_UNKEX 3

id_t resolve_controlfs (file_t device, char *path, char **newpath)
{
  if(!strcmp(path, "restart")) {
    return CTLFS_RESTART;
  } else if(!strcmp(path, "kex")) {
    return CTLFS_KEX;
  } else if(!strcmp(path, "unkex")) {
    return CTLFS_UNKEX;
  } else {
    return -1;
  }
}

int write_controlfs (id_t id, fsize_t start, fsize_t size, void *buffer, file_t device)
{
  switch (id) {
  case CTLFS_KEX:
    return load_kex (buffer);
    break;
    
  case CTLFS_UNKEX:
    return unload_kex (buffer);
    break;
    
  default:
    return -1;
    break;
  }
}

int unlink_controlfs(id_t id, file_t device)
{
  switch(id) {
  case CTLFS_RESTART:
    restart_system();
    break;
  default:
    return -1;
    break;
  }
}



void init_controlfs()
{
  vfs_handler_t *controlfs_handler;

  controlfs_handler = calloc(sizeof(vfs_handler_t), 1);
  
  //  controlfs_handler->read_func = &read_controlfs;
  controlfs_handler->write_func = &write_controlfs;
  controlfs_handler->resolve_func = &resolve_controlfs;
  controlfs_handler->unlink_func = &unlink_controlfs;
  //  controlfs_handler->creat_func = &creat_controlfs;
  //  controlfs_handler->stat_func = NULL;

  add_vfs_handler (VFS_CONTROLFS, controlfs_handler);
}
