/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

#include <tyros.h>
#include <ntx.h>
#include <ntxbase.h>
#include <vfs.h>
#include <filesystems.h>
#include <devfs.h>

extern int terminal_owner;

int init_devices(nsentry_t *dev, int nohdd, int nofdd)
{
nsentry_t *ramdisk, *vc, *null, *zero;

//init standard devices (ramdisk, null, zero & virtual console)
ramdisk = add_nsentry(dev, create_nsentry(NSE_OBJECT), "ramdisk");
vc = add_nsentry(dev, create_nsentry(NSE_OBJECT), "vc");
zero = add_nsentry(dev, create_nsentry(NSE_OBJECT), "zero");
null = add_nsentry(dev, create_nsentry(NSE_OBJECT), "null");

ramdisk->nsstatus.vfsid = VFS_DEVFS;
ramdisk->nsstatus.ident = DEVFS_IDENT_RAMDISK;
ramdisk->nsstatus.size = hal.ramdisk_size;

vc->nsstatus.vfsid = VFS_DEVFS;
vc->nsstatus.ident = DEVFS_IDENT_CONSOLE;
vc->nsstatus.size = 0x1000;

null->nsstatus.vfsid = VFS_DEVFS;
null->nsstatus.ident = DEVFS_IDENT_NULL;
null->nsstatus.size = 0x1000;

zero->nsstatus.vfsid = VFS_DEVFS;
zero->nsstatus.ident = DEVFS_IDENT_ZERO;
zero->nsstatus.size = 0x1000;
if(!nofdd)
 init_fdd(dev);
if(!nohdd)
 init_hdd(dev);
 
init_rs232(dev);
}

int stat_devfs (int id, ntx_stat_t *buffer, file_t device)
{
  printf("statting devfs\n");
  return -1;
}

//int read_hdd (int atanum, int devnum, fsize_t start, fsize_t size, void *buffer);

int extract_major (unsigned int id)
{
  return (id & 0x0000FFFF);
}

int extract_minor (unsigned int id)
{
  return (id & 0xFFFF0000) >> 16;
}

int get_ident (unsigned int major, unsigned int minor)
{
  return ((minor << 16) | (major));
}

int read_devfs (id_t lid, fsize_t start, fsize_t size, void *buffer, file_t device)
{
  unsigned int major, minor, i;
  unsigned int id = (unsigned int) lid;
  int* buf = buffer;

  major = extract_major(id);
  minor = extract_minor(id);

printf("read_devfs: reading %i\n", major);
  
  switch(major) {
  case DEVFS_IDENT_RAMDISK:
    memcpy(buffer, hal.ramdisk_binary + (unsigned int)start, (unsigned int) size);
    return size;
    break;

  case DEVFS_IDENT_CONSOLE:
    terminal_owner = tasks.current;
    return getsn(buffer, size);
    break;
  
  case DEVFS_IDENT_NULL:
    return 0;
    break;

  case DEVFS_IDENT_ZERO:
    for(i=0; i<size; i++)
     buf[i] = 0;
    return size;
    break;
    
  default:
    if(is_blkdev(major) == 0)
     return blkdev_read (major, minor, start, size, buffer);
    return -1;
    break;
  } 
}

int write_devfs (id_t lid, fsize_t start, fsize_t size, void *buffer, file_t device)
{
  int id_type;
  int id_device;
  int i;
  char *buff;
  unsigned int id = (unsigned int) lid;

  buff = buffer;
  
  switch(id) {
  case DEVFS_IDENT_CONSOLE:
    terminal_owner = tasks.current;
    for(i = 0; i < size; i++) {
      putchar(buff[i]);
    }
    return size;
    break;
    
  case DEVFS_IDENT_RAMDISK:
    memcpy(hal.ramdisk_binary + (unsigned int) start, buffer, (unsigned int) size);
    return size;
    break;

  case DEVFS_IDENT_ZERO:
  case DEVFS_IDENT_NULL:
    return size;
    break;
    
  default:
    return -1;
    break;
  }
}


int unlink_devfs (int id, file_t device)
{
  return -1;
}

void init_devfs()
{
  vfs_handler_t *devfs_handler;
  
  devfs_handler = calloc(sizeof(vfs_handler_t), 1);

  devfs_handler->read_func = &read_devfs;
  devfs_handler->write_func = &write_devfs;
  devfs_handler->stat_func = &stat_devfs;
  devfs_handler->unlink_func = &unlink_devfs;
  
  add_vfs_handler (VFS_DEVFS, devfs_handler);
}
