/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

/*
  FAT12 driver
  
  i know this code is awful... 

  NOTE:
  FAT and root directory are completely cached (they are never too big in FAT12)
*/

#include <tyros.h>
#include <ntx.h>
#include <ntxbase.h>
#include <vfs.h>
#include <filesystems.h>
#include <assert.h>

void transform_filename(char *filename, char *realname);
char upper (char c);
int is_divisor(char c);

typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;

#define FAT_ROOTDIR -2

typedef struct fat_info_struct {
  char oem_name[0x10];
  int bytes_per_sec;
  int sec_per_clust;
  int reserved_sectors;
  int numfats;
  int root_entries;
  int total_sectors;
  int media;
  int fat_size;
  int sec_per_track;
  int num_heads;
  int hidden_sectors;
  int total_sectors32;
} fat_info_t;

typedef struct fat12_cache_struct {
  char *root;
  char *fat;
  fat_info_t fatinfo;
} fat12_cache_t;

int find_free_cluster12 (fat12_cache_t *cache, int except);

/*
  seeks the cluster which containts the specified file position
*/
int seek_cluster12 (int cluster, fsize_t start, int *cloff, fat_info_t *bpb, char *fat, file_t device)
{
  unsigned short wren, newwren;//word-sized fat entry
  int cluster_size;
  int noc;
  int off;
  int i;

  if(start == 0) {
    *cloff = 0;
    return cluster;
  } else {
    
    cluster_size = bpb->bytes_per_sec*bpb->sec_per_clust;

    noc = ((int)start)/cluster_size;

    if(noc == 0) {
      *cloff = start;
      return cluster;
    }

    off = ((int)start)%cluster_size;

    //    printf("Seeking %X (%i clusters, off = %i)\n",(unsigned int)start, noc, off);

    wren = cluster;

    for(i = 0; i < noc; i++) {
      
      
      newwren = * (unsigned short *) (fat + (wren*3)/2);
      if( (wren & 1) == 0) {
	newwren &= 0x0FFF;
      } else {
	newwren >>= 4;
      }
      wren = newwren;
      //      printf("newwren: %X.... \n", wren);
    
      if(wren >= 0xFF8) {
	//	printf("EOF reached!\n");
	return -1;
      }
    }

    
    //    printf("Seeking cluster %X - %X, WREN: %X, cluster size: %i\n", cluster, (int) start, wren, cluster_size);
    *cloff = off;
    return wren;
  }

}

int in_rootdir(int id, fat12_cache_t *cache)
{
  int rootdir_start, rootdir_end;

  rootdir_start = (cache->fatinfo.reserved_sectors + cache->fatinfo.numfats*cache->fatinfo.fat_size)*512;
  rootdir_end = rootdir_start + cache->fatinfo.root_entries*32;

  if(id >= rootdir_start && id < rootdir_end) {
    return 1;
  } else {
    return 0;
  }
}

int load_cluster_chain12 (int cluster, int cloff, void *buffer, int size, int filesize, fat_info_t *bpb, char *fat, file_t device)
{
  unsigned short wren, newwren;
  int i;
  int cluster_size;
  int noc;
  char *localbuf;
  int lsize;
  int rsize;

  cluster_size = bpb->bytes_per_sec*bpb->sec_per_clust;
  noc = ((int)size)/cluster_size + 1;
    
  //  printf("lcc12: cloff: %i, realsize = %i, noc = %i\n", cloff, filesize, noc);
  wren = cluster;

  
  if(cloff > filesize) {
    return 0;
  }

  for(i = 0; i < noc; i++) {

    newwren = * (unsigned short *) (fat + (wren*3)/2);
    if( (wren & 1) == 0) {
      newwren &= 0x0FFF;
    } else {
      newwren >>= 4;
    }
    
  
    if(i == 0 && cloff != 0) {
      copy_cluster12 (wren, localbuf, bpb, fat, device);
      strcpy(buffer, localbuf + cloff, cluster_size - cloff);
      buffer = (buffer + cloff) - 512;

      if(noc == 1) {
	return (cluster_size - cloff);
      }


    } else {
     
      
      if(i == noc - 1) {
	//      printf("caching %X through local buffer\n", wren);
	lsize = size - (noc - 1)*cluster_size;
	//      printf("local size: %i (total: %i, clusters read: %i)\n", lsize, size, noc - 1);
	
	localbuf = alloc_pmem(mem_to_pages(cluster_size));
	copy_cluster12 (wren, localbuf, bpb, fat, device);
	memcpy(buffer + i*cluster_size, localbuf, lsize);
	free_pmem(localbuf);
	//      printf("Total read: %i\n", i*cluster_size + lsize);
	
	rsize = i*cluster_size + lsize;
	
	if(rsize > filesize) {
	  return filesize;
	} else {
	  return rsize;
	}
	
      } else {
	copy_cluster12 (wren, buffer + i*cluster_size, bpb, fat, device);
      }
      
    }
    
    wren = newwren;
    
    if(newwren >= 0xFF8) {
      //      printf("EOF reached\n");
      return filesize - cloff;
      //      return i*cluster_size;
    }
    
  }  
  
  //  return load_cluster_chain12(wren, buffer + 512, bpb, fat, device) + 512;
}

int copy_cluster12 (int cluster, void *buffer, fat_info_t *bpb, char *fat, file_t device)
{
  int sector;
  int cluster_size;

  sector = bpb->reserved_sectors + bpb->numfats*bpb->fat_size + (bpb->root_entries*32)/512 + (cluster - 2)*bpb->sec_per_clust;
  
  cluster_size = bpb->sec_per_clust*bpb->bytes_per_sec;

  //  printf("Target sector: %X -> %X\n", sector, buffer);

  read_file(device, sector*512, cluster_size, buffer);

}

int stat_fat12 (id_t lid, ntx_stat_t *buffer, file_t device)
{
  char dirent[0x30];
  unsigned int id = (unsigned int) lid;

  if(id == FAT_ROOTDIR) {
    buffer->st_size = 0;
    buffer->st_mode = ST_IFDIR;
    return 0;
  }

  read_file(device, id, 32, dirent);

  //  printf("stat_fat12 (id: %X)\n", id);

  buffer->st_size = * (int *) &dirent[0x1C];
  buffer->st_mode = 0;
  
  //  printf("File size: %i\n", buffer->st_size);

  return 0;
}

int reverse_transform_filename (char *realname, char *filename)
{
  int i;
  
  for(i = 0; i < 11; i++) {
    if(i == 8) {
      if(realname[i] != ' ') {
	*filename = '.';
	filename++;
      }
    }

    if(realname[i] != ' ') {
      *filename = realname[i];
      filename++;
    }
  }
  *filename = 0;
}

int readdir_fat12 (id_t lid, fsize_t start, fsize_t size, char *buffer,
		   fat_info_t *bpb, char *fat, file_t device)
{
  unsigned char *directory;
  char filename[0x10];
  int rootdir_sector;
  int rootdir_size;
  int data_sector;
  int dirents; //number of directory entries
  int i;
  int result;
  char *l;
  unsigned int id = (unsigned int) lid;

  if(id == FAT_ROOTDIR) {
    rootdir_sector = bpb->reserved_sectors + bpb->fat_size*bpb->numfats;
    data_sector = rootdir_sector+(32*bpb->root_entries)/512;
    dirents = bpb->root_entries;
    
    rootdir_size = 32*dirents;
    
    directory = alloc_pmem(mem_to_pages(rootdir_size));
    read_file(device, rootdir_sector*512, rootdir_size, directory);

    l = buffer;

    for(i = 0; i < dirents; i++) {
      
      if(*(directory + i*32) == 0) {
	break;
      }
      
      if(*(directory + i*32) == 0xE5) {
	continue;
      }
      
      reverse_transform_filename (directory + i*32, filename);
      //      printf("found file: %s\n", filename);
      l += sprintf(l,"%s", filename);
      *l = 0;
      l++;
    }
    
    

  } else {
    printf("readdir_fat12: non-root directory!\n");
    for(;;);
  }

  return (l - buffer);
}

//returns first cluster of the file
int grow_fat12(id_t lid, int oldsize, int newsize, fat12_cache_t *cache, file_t device)
{
  char dirent[0x30];
  int root_start;
  int newclusters;//count of new clusters
  int cluster_size;
  int i;
  int newc;//new cluster
  int lastc;//last cluster
  int cloff;
  int cluster;
  int prev;
  int id = (unsigned int) lid;

  cluster_size = cache->fatinfo.sec_per_clust*512;

  newclusters = (newsize/cluster_size) - (oldsize/cluster_size);

  if(newclusters == 0){
    
    if(in_rootdir(id, cache)) {
      root_start = (cache->fatinfo.reserved_sectors + cache->fatinfo.numfats*cache->fatinfo.fat_size)*512;
    
      memcpy(dirent, cache->root + (id - root_start), 32);
      
      * (unsigned int *) &dirent[0x1C] = (unsigned int) newsize;
      
      memcpy(cache->root + (id - root_start), dirent, 32);

      write_file(device, id, 0x20, dirent);

      return * (short *) &dirent[0x1A];
      
    } else {
      printf("grow_fat12: non-root directory\n");
      return -1;
    }
    
  } else {
    printf("growing by %i clusters\n", newclusters);
    
    root_start = (cache->fatinfo.reserved_sectors + cache->fatinfo.numfats*cache->fatinfo.fat_size)*512;
    
    memcpy(dirent, cache->root + (id - root_start), 32);
    
    cluster = * (short *) &dirent[0x1A];

    
    lastc = seek_cluster12(cluster, oldsize, &cloff, &cache->fatinfo, cache->fat, device);
    
    //    printf("lastc = %X\n", lastc);
    
    for(i = 0; i < newclusters; i++) {
      newc = find_free_cluster12(cache, lastc);
      //      printf("Updating %X with %X\n", lastc, newc);
      prev = update_fat12_entry(lastc, newc, cache, device);
      //      printf("fat %X = %X, prev = %X\n", lastc, newc, prev);
      
      lastc = newc;
    }
    
    update_fat12_entry(lastc, 0x0FF8, cache, device);

    * (unsigned int *) &dirent[0x1C] = (unsigned int) newsize;
    memcpy(cache->root + (id - root_start), dirent, 32);

    write_file(device, id, 0x20, dirent);
    
    return cluster;
  }
}


int write_cluster12 (int cluster, int offset, int size, void *buffer, fat12_cache_t *cache,
		     file_t device)
{
  int sector;
  int cluster_size;
  fat_info_t *bpb;
  char *localbuf;

  bpb = &cache->fatinfo;
  
  sector = bpb->reserved_sectors + bpb->numfats*bpb->fat_size + (bpb->root_entries*32)/512 + (cluster - 2)*bpb->sec_per_clust;
  
  cluster_size = bpb->sec_per_clust*bpb->bytes_per_sec;
  
  //  printf("writing cluster12, cluster:%i, offset:%i, size:%i, sector: %X\n", cluster, offset, size, sector);
  
  if(cluster_size == size) {
    write_file (device, sector*512, cluster_size, buffer);
  } else {
    localbuf = alloc_pmem(mem_to_pages(cluster_size));
    read_file(device, sector*512, cluster_size, localbuf);
    memcpy(localbuf + offset, buffer, size);
    write_file(device, sector*512, cluster_size, localbuf);
    free_pmem(localbuf);
  }
}

int write_fat12 (id_t lid, fsize_t start, fsize_t size, void *buffer, file_t device)
{
  ntx_stat_t statbuf;
  fat12_cache_t *cache;
  int cluster;
  int tc;//target cluster
  int ts;//target sector
  char *cl;
  int cloff, clleft;
  int i;
  int cluster_size;
  int cluster_start,clusters_count;
  char dirent[0x30];
  unsigned int istart = (unsigned int) start, isize = (unsigned int) size;
  unsigned int coffset, csize, cbuffer, cno;
  unsigned int hb;
  unsigned int id = (unsigned int) lid;

  printf("writing fat12, start: %i, size: %i, buffer: %X\n", (unsigned int) start, (unsigned int) size, buffer);
  stat_fat12(id, &statbuf, device);
  printf("file size: %i bytes\n", statbuf.st_size);
    
  cache = access_device_cache(device);
  assert(cache != NULL);

  if(start + size > statbuf.st_size) {
    cluster = grow_fat12(id, statbuf.st_size, start + size, cache, device);
    if(cluster == -1){
      return -1;
    }
  } else {
    read_file(device, id, 0x20, dirent);
    cluster = * (short *) &dirent[0x1A];
  }
  
  cluster_size = 512*cache->fatinfo.sec_per_clust;

  
  
  cluster_start = istart/cluster_size;
  
  clusters_count = ((istart%cluster_size + isize)/cluster_size) + 1;

  printf("cluster_start: %i, clusters_count: %i\n", cluster_start, clusters_count);


  for(cno = cluster_start; cno < cluster_start + clusters_count; cno++) {

    if(cno*cluster_size >= istart) {
      coffset = 0;
    } else {
      coffset = istart - cno*cluster_size;
    }
    
    if(coffset > 0) {
      cbuffer = buffer;
    } else {
      cbuffer = buffer + (cno*cluster_size - istart);
    }
    
    if( (cno + 1)*cluster_size <= (istart + isize)) {
      hb = cluster_size;
    } else {
      hb = cluster_size - ((cno + 1)*cluster_size - (istart + isize));
    }
    csize = hb - coffset;

    
    tc = seek_cluster12(cluster, cno*cluster_size, &cloff, &cache->fatinfo, cache->fat, device);
    
    write_cluster12(tc, coffset, csize, cbuffer, cache, device);
  }

  

  return size;
  
}



int read_fat12 (id_t lid, fsize_t start, fsize_t size, void *buffer, file_t device)
{
  char superblock[512];
  char dirent[0x30];
  int fat_size;
  char *fat;
  fat_info_t bpb;
  unsigned short wren;
  int fclust;
  int cloff; //offset in cluster
  int result;
  int cluster;
  int cluster_size;
  int filesize;
  ntx_stat_t statbuf;
  unsigned int id = (unsigned int) lid;

  stat_fat12(id, &statbuf, device);

  filesize = statbuf.st_size;
  //  printf("Read_fat12: id: %X, start: %X\n", id, (int)start);


  read_file(device, 0, 512, superblock);

  fill_fat_info(&bpb, superblock);

  cluster_size = bpb.bytes_per_sec*bpb.sec_per_clust;

  fat_size = bpb.fat_size * 512;

  fat = alloc_pmem(mem_to_pages(fat_size));

  load_fat ( &bpb, fat, fat_size, device);


  if(id == FAT_ROOTDIR) {
    return readdir_fat12 (FAT_ROOTDIR, start, size, buffer, &bpb, fat, device);
  } else {
    read_file(device, id, 32, dirent);
    cluster = * (short *) &dirent[0x1A];
  }
  
  fclust = seek_cluster12 (cluster, start,&cloff, &bpb, fat, device);

  if(fclust == -1) {
    result = 0;
  } else {
    filesize -= ( ((unsigned int)start/cluster_size)*cluster_size );
    result = load_cluster_chain12 (fclust, cloff, buffer, size, filesize, &bpb, fat, device);
  }
  
  free_pmem(fat);

  //  printf("%i bytes read\n", result);

  return result;
}


int split_filename (char *path, char **arglist, int *argcount)
{
  char *buffer;
  int buflen;
  int i;
  int last;
  char **argv;
  int argc;
  
  argv = arglist;

  buflen = strlen(path) + 1;
  buffer = path;

  argc = 0;
  last = 0;

  for(i = 0; i < buflen; i++) {
      
    if(is_divisor(buffer[i])) {
      
      buffer[i] = 0;
      
      argv[argc] = &buffer[last];
      
      argc++;
      
      last = i + 1;
      while(is_divisor(buffer[last])){
	last++;
	i++;
      }
      
    }
    
  }

  argv[argc] = NULL;
  *argcount = argc;
  return 0;
}


int get_cluster (int dir_cluster, char *filename, int *dirent, fat_info_t *bpb, char *fat, fat12_cache_t *cache, file_t device)
{
  char *directory;
  int dir_sector;
  char realname[0x10];
  int rootdir_sector;
  int rootdir_size;
  int data_sector;
  int dirents; //number of directory entries
  int i;
  int result;

  

  transform_filename(filename, realname);
  if(dir_cluster == -1) {
    //root directory
    directory = cache->root;
    dirents = cache->fatinfo.root_entries;
    dir_sector = cache->fatinfo.reserved_sectors + cache->fatinfo.fat_size*cache->fatinfo.numfats;
  
  } else {
    printf("Non-root get_cluster not supported!\n");
    for(;;);
  }
  
  for(i = 0; i < dirents; i++) {
    if(!memcmp(realname, directory + i*32, 10)) {
      *dirent = dir_sector*512 + i*32;
      result = * (short *) (directory + i*32 + 0x1A);
      break;
    }
    
    if(*(directory + i*32) == 0xE5) {
      continue;
    }

    if(*(directory + i*32) == 0) {
      result = -1;
      break;
    }
  }
  
  return result;
}

unpack_directory (file_t directory, file_t *ndevice, id_t *dirid)
{
  file_t device;
  int i;
  
  device = directory;
  i = 0;
  do {
    i++;
  } while (directory.id[i] != -1);
  device.id[i - 1] = -1;
  *dirid = directory.id[i - 1];
  
  *ndevice = device;
  
  return 0;
}

int get_free_dirent (int dirid, fat12_cache_t *cache, file_t device)
{
  char *root;
  int root_start;
  int dirents;
  int i;


  
  if(dirid == FAT_ROOTDIR) {
    root_start = (cache->fatinfo.reserved_sectors + cache->fatinfo.numfats*cache->fatinfo.fat_size)*512;
    root = cache->root;
    dirents = cache->fatinfo.root_entries;
    
    
    for(i = 0; i < dirents; i++) {
      if(root[i*32] == 0xE5 || root[i*32] == 0x00) {
	return root_start + i*32;
      }
    }
    
    return -1;
  } else {
    printf("creating files in non-root directory not supported\n");
    return -1;
  }
  

  for(;;);
}

int find_free_cluster12 (fat12_cache_t *cache, int except)
{
  unsigned short wren;
  int i;
  int numclusters;
  char *fat = cache->fat;
  
  numclusters = cache->fatinfo.total_sectors/cache->fatinfo.sec_per_clust;
  
  for(i = 2; i < numclusters; i++) {
    wren = * (unsigned short *) (fat + (i*3)/2);
    if( (i & 1) == 0) {
      wren &= 0x0FFF;
    } else {
      wren >>= 4;
    }
    
    if(wren == 0x00 && i != except) {
      return i;
    }
    
  }
  
  return -1;

}
int creat_fat12 (char *file, int mode, file_t directory)
{
  file_t device;
  int dirid;
  int newid;
  fat12_cache_t *cache;
  char dirent[0x30];
  int cluster;
  char realname[0x10];
  int root_start;

  printf("creat_fat12\n");

  unpack_directory(directory, &device, &dirid);

  cache = access_device_cache(device);
  
  assert(cache != NULL);

  if(dirid == FAT_ROOTDIR) {
    
    newid = get_free_dirent (dirid,cache, device);

    printf("get_free_dirent returned %X\n", newid);
    
    cluster = find_free_cluster12(cache, -1);

    printf("free cluster: %X\n", cluster);

    if(cluster == -1) {
      printf("no free space!\n");
      return -1;
    }
    
    transform_filename(file, realname);
    
    memcpy(dirent, realname, 11);
    
    dirent[0x0B] = 0; //attribute
    dirent[0x0C] = 0; //reserved for NT
    dirent[0x0D] = 0; //FAT32 only
 
    * (short *) &dirent[0x0E] = 0; //FAT32 only
    * (short *) &dirent[0x10] = 0; //FAT32 only
    * (short *) &dirent[0x12] = 0; //FAT32 only
    * (short *) &dirent[0x14] = 0; //FAT32 only
    * (short *) &dirent[0x16] = 0; //last write time
    * (short *) &dirent[0x18] = 0; //last write date
    * (short *) &dirent[0x1A] = cluster; //first cluster
    * (int *) &dirent[0x1C] = 0; //file size

    write_file(device, newid, 32, dirent);
    
    update_fat12_entry (cluster, 0xFF8, cache, device);

    root_start = (cache->fatinfo.reserved_sectors + cache->fatinfo.numfats*cache->fatinfo.fat_size)*512; 
    
    memcpy(cache->root + (newid - root_start), dirent, 32);
    
  } else {
    printf("creating files in non-root directory not supported\n");
    return -1;
  }
}


//func returns previous value of the fat entry
int update_fat12_entry (unsigned int cluster, int new, fat12_cache_t *cache, file_t device)
{
  char *fat;
  unsigned short wren;
  unsigned short prevv, prevf;
  unsigned short newf;
  int fat_start;
  int wst;

  fat = cache->fat;
  
  wren = (cluster*3)/2;

  prevf = * (short *) (fat + wren);
  
  if(cluster & 1) {
    prevv = (prevf & 0xFFF0) >> 4;
    newf = (prevf & 0x000F) | (new << 4);
  } else {
    prevv = prevf & 0x0FFF;
    newf = (prevf & 0xF000) | new;
  }
  
  

  
  * (short *) (fat + wren) = newf;
  
  fat_start = cache->fatinfo.reserved_sectors*512;
  
  wst = (wren/512)*512;
  
  //  printf("updating fat12: cluster: %X, wren: %X, prev: %X, wst: %X\n", cluster, wren, prevv, wst);

  write_file(device, fat_start + wst, 512, fat + wst);
  
  return prevv;

}

int unlink_fat12(id_t lid, file_t device)
{
  unsigned char dirent[0x30];
  fat12_cache_t *cache;
  int fclust;
  int prev;
  int rootdir_start;
  int id = (int)lid;

  //  printf("fat12 unlinking\n");
  
  if(id == FAT_ROOTDIR) {
    return -1;
  }
  
  cache = access_device_cache(device);
  
  //it should be impossible to unlink file without resolving it first
  //(and resolving will reload the cache if it is needed)
  assert(cache != NULL);
  
  
  read_file(device, id, 32, dirent);
  
  fclust = * (short *) &dirent[0x1A];
  
  prev = fclust;

  do{
    prev = update_fat12_entry(prev, 0x000, cache, device);
  } while(prev < 0xFF8); 
  
  dirent[0] = 0xE5;
  write_file(device, id, 32, dirent);

  if(in_rootdir(id, cache)) {
    //    printf("updating cache...\n");
    rootdir_start = (cache->fatinfo.reserved_sectors + cache->fatinfo.numfats*cache->fatinfo.fat_size)*512; 
    cache->root[id - rootdir_start] = 0xE5;
  }
  
  return 0;
}

void *init_fat12_caching(file_t device)
{
  fat12_cache_t *cache;
  char superblock[0x210];
  fat_info_t fatinfo;
  int fat_size;
  int root_size;
  char *fat;
  char *root;

  cache = alloc_device_cache(device, sizeof(fat12_cache_t));
  
  read_file (device, 0, 512, superblock);
  
  fill_fat_info(&cache->fatinfo, superblock);
  
  fat_size = cache->fatinfo.fat_size * 512;
  fat = alloc_pmem(mem_to_pages(fat_size));
  load_fat(&cache->fatinfo, fat, fat_size, device);
  cache->fat = fat;
  
  root_size = 32*cache->fatinfo.root_entries;
  root = alloc_pmem(mem_to_pages(root_size));
  load_root(&cache->fatinfo, root, root_size, device);
  cache->root = root;

  return cache;
}


int load_root (fat_info_t *bpb, char *root, int size, file_t device)
{
  int root_start;

  root_start = (bpb->reserved_sectors + bpb->numfats*bpb->fat_size)*512;
  //  printf("root start: %X, size: %X\n", root_start, size);
  read_file(device, root_start, size, root);
  
  return 0;
}


void cleanup_fat12_caching (file_t device)
{
  
}

id_t resolve_fat12 (file_t device, char *path, char **newpath)
{
  char *arglist[0x100];
  int argcount;
  int last;
  int i;
  fat_info_t *bpb;
  char oemname[0x10];
  int next_cluster;
  char *fat;
  char *root;
  int dirent;
  char npath[PATH_MAX];
  fat12_cache_t *cache;

  
  if(!strcmp(path,"")) {
    return FAT_ROOTDIR;
  }

  strncpy(npath, path, PATH_MAX);

  split_filename (npath, arglist, &argcount);

  /*  read_file(device, 0, 512, superblock);

  fill_fat_info(&bpb, superblock);
  
  fat_size = bpb.fat_size * 512;

  fat = alloc_pmem(mem_to_pages(fat_size));

  load_fat ( &bpb, fat, fat_size, device);
  */
  
  cache = access_device_cache(device);

  if(cache == NULL) {
    cache = init_fat12_caching(device);
  }
  
  
  fat = cache->fat;
  root = cache->root;
  bpb = &cache->fatinfo;
  

  next_cluster = -1;
  for(i = 0; i < argcount; i++) {
    next_cluster = get_cluster(next_cluster, arglist[i],&dirent, bpb, fat, cache, device);
    if(next_cluster == -1) {
      //printf("File not found! (%s)\n", path);
      dirent = -1;
    }
  }
  
  //  printf("fat12 resolve result: %X (path: %s)\n", dirent, path);
  return (id_t) dirent;
}

int load_fat (fat_info_t *bpb, char *fat, int size, file_t device)
{
  int fat_start;

  fat_start = bpb->reserved_sectors*512;
  
  read_file(device, fat_start, size, fat);

}

void transform_filename(char *filename, char *realname)
{
  int i, j;
  char *st;
  
  st = realname;
  
  for(i = 0; i < 8; i++) {
    if(*filename != '.' && *filename != 0) {
      *realname = upper(*filename);
      filename++;
    } else if(*filename == 0) {
      *realname = ' ';
      while(i < 11) {
	*realname = ' ';
	realname++;
	i++;
      }

    } else if(*filename == '.') {
      while(i < 8) {
	*realname = ' ';
	realname++;
	i++;
      }
      filename++;
      for(j = 0; j < 3; j++) {
	if(*filename != 0) {
	  *realname = upper(*filename);
	  filename++;
	} else {
	  *realname = 0;
	}
	realname++;
      }
    }
    realname++;
  }
  
  *realname = 0;
}

char upper (char c)
{
  if(c >= 'a' && c <= 'z') {
    return ('A' + (c - 'a'));
  } else {
    return c;
  }
}


int fill_fat_info(fat_info_t *fatinfo, char *superblock)
{
  memcpy(fatinfo->oem_name, &superblock[0x03], 8);
  fatinfo->oem_name[8] = 0;
  fatinfo->bytes_per_sec = * (short *) &superblock[0x0B];
  fatinfo->sec_per_clust = superblock[0x0D];
  fatinfo->reserved_sectors = * (short *) &superblock[0x0E];
  fatinfo->numfats = superblock[0x10];
  fatinfo->root_entries = * (short *) &superblock[0x11];
  fatinfo->total_sectors = * (short *) &superblock[0x13];
  fatinfo->media = superblock[0x15];
  fatinfo->fat_size = * (short *) &superblock[0x16];
  fatinfo->sec_per_track = * (short *) &superblock[0x18];
  fatinfo->num_heads = * (short *) &superblock[0x1A];
  fatinfo->hidden_sectors = * (int *) &superblock[0x1C];
  fatinfo->total_sectors32 = * (int *) &superblock[0x20];
  
}

int is_divisor(char c)
{
  if(c == '/' || c == '\n' || c == 0) {
    return 1;
  } else {
    return 0;
  }
}



void init_fatfs()
{
  vfs_handler_t *fatfs_handler;
  
  fatfs_handler = calloc(sizeof(vfs_handler_t), 1);
  
  fatfs_handler->name = "fat12";

  fatfs_handler->read_func = &read_fat12;
  fatfs_handler->write_func = &write_fat12;
  fatfs_handler->resolve_func = &resolve_fat12;
  fatfs_handler->stat_func = &stat_fat12;
  fatfs_handler->unlink_func = &unlink_fat12;
  fatfs_handler->creat_func = &creat_fat12;

  add_vfs_handler (VFS_FAT12, fatfs_handler);
}
