/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

#include <tyros.h>
#include <ntx.h>
#include <ntxbase.h>
#include <vfs.h>
#include <filesystems.h>
#include <infofs.h>
#include <core.h>

char *infofs_file_version = INFOFS_FILE_VERSION;
char infofs_file_freemem[0x20];
char infofs_file_memmap[0x1000];


extern free_memory;

int read_infofs (id_t lid, fsize_t start, fsize_t size, void *buffer, file_t device)
{
  int i;
  int num;
  char argbuf[0x1000];
  int id = (int) lid;
  char *p = &argbuf;

  switch(id) {

  case INFOFS_IDENT_WM:
    return 0;
    break;

  case INFOFS_IDENT_VERSION:
    memcpy(buffer, infofs_file_version + (unsigned int) start, (unsigned int) size);
    
    
    num = strlen(infofs_file_version + (unsigned int) start);
    
    if((unsigned int) size < num) {
      return (unsigned int) size;
    } else {
      return num;
    }

    break;
    
  case INFOFS_IDENT_FREEMEM:
    sprintf(infofs_file_freemem, "%i\n", free_memory);
    
    memcpy(buffer, infofs_file_freemem + (unsigned int) start, (unsigned int) size);
    
    
    num = strlen(infofs_file_freemem + (unsigned int) start);
    
    if((unsigned int) size < num) {
      return (unsigned int) size;
    } else {
      return num;
    }
    break;

  case INFOFS_IDENT_MEMMAP:
    display_memory_map(infofs_file_memmap);

    memcpy(buffer, infofs_file_memmap + (unsigned int) start, (unsigned int) size);
    
    
    num = strlen(infofs_file_memmap + (unsigned int) start);
    
    if((unsigned int) size < num) {
      return (unsigned int) size;
    } else {
      return num;
    }
    
    break;
    
    break;

  case INFOFS_IDENT_PROCESSES:
    for(i = 0; i < tasks.count; i++) {
      p += sprintf(p, "%i ", tasks.queue[i]);
    }
    return read_mem (argbuf, strlen(argbuf) + 1, (unsigned int) start, (unsigned int) size, buffer);
    break;

  case INFOFS_IDENT_FILESYSTEMS:
    for(i = 0; i < vfs_registry->nhandlers; i++) {
      if(vfs_registry->vfs_handler[i] != NULL) {
	if(vfs_registry->vfs_handler[i]->name != NULL) {
	  p += sprintf(p, "%s\n", vfs_registry->vfs_handler[i]->name);
	}
      }
    }
    return read_mem (argbuf, strlen(argbuf) + 1, (unsigned int) start, (unsigned int) size, buffer);    
    break;

  case INFOFS_IDENT_EXTENSIONS:
    if(kex_registry->count == 0) {
      p += sprintf(p, "# no loaded extensions\n");
    } else {
      for(i = 0; i < kex_registry->count; i++) {
	p += sprintf(p, "Name '%s' File: '%s'\n",
		     kex_registry->kex[i]->name,
		     kex_registry->kex[i]->filename);
	
      }
    }
    return read_mem (argbuf, strlen(argbuf) + 1, (unsigned int) start, (unsigned int) size, buffer);
    break;
    
  default:
    return -1;
    break;
  }
}

id_t resolve_infofs (file_t device, char *path, char **newpath)
{
  if(!strcmp(path, "version")) {
    return INFOFS_IDENT_VERSION;
  } else if(!strcmp(path,"freemem")) {
    return INFOFS_IDENT_FREEMEM;
  } else if(!strcmp(path, "filesystems")) {
    return INFOFS_IDENT_FILESYSTEMS;
  } else if(!strcmp(path, "extensions")) {
    return INFOFS_IDENT_EXTENSIONS;
  } else if(!strcmp(path, "memmap")) {
    return INFOFS_IDENT_MEMMAP;
    
  } else if(!strcmp(path, "processes")) {
    return INFOFS_IDENT_PROCESSES;
  
  } else if(!strcmp(path, "wm")) {

    if(hal.mode != 0) {
      return INFOFS_IDENT_WM;
    } else {
      return -1;
    }
  } else {
    return -1;
  }
}

//infofs is read-only
int write_infofs (id_t id, fsize_t start, fsize_t size, void *buffer, file_t device)
{
  return -1;
}

int unlink_infofs(id_t id, file_t device)
{
  return -1;
}

int creat_infofs(char *path, int mode, file_t device)
{
  return -1;
}

void init_infofs()
{
  vfs_handler_t *infofs_handler;

  infofs_handler = calloc(sizeof(vfs_handler_t), 1);
  
  infofs_handler->read_func = &read_infofs;
  infofs_handler->write_func = &write_infofs;
  infofs_handler->resolve_func = &resolve_infofs;
  infofs_handler->unlink_func = &unlink_infofs;
  infofs_handler->creat_func = &creat_infofs;
  infofs_handler->stat_func = NULL;

  add_vfs_handler (VFS_INFOFS, infofs_handler);


}
