/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */


#include <tyros.h>
#include <ntx.h>
#include <core.h>
#include <ntxbase.h>
#include <vfs.h>
#include <filesystems.h>

nsentry_t *add_static_nsentry (nsentry_t *target, char *name, char *contents)
{
  nsentry_t *new;

  new = create_nsentry(NSE_OBJECT);
  
  new->nsstatus.vfsid = VFS_MEMFS;
  new->nsstatus.ident = contents;
  new->nsstatus.size = strlen(contents);
  
  return  add_nsentry(target, new, name);
}

int read_memfs (id_t lid, fsize_t start, fsize_t size, void *buffer, file_t device)
{
  int realsize;
  char *id = (char *) lid;

  realsize = strlen(id);
  
  //  printf("readmemfs: realsize = %i, start = %i\n", realsize, start);

  if(start >= realsize) {
    return 0;
  }
  
  if( (size + start) > realsize) {
    size = realsize - start;
  }
  
  if(size < 0 || size == 0) {
    return 0;
  }

  strncpy(buffer, id + start, size);

  return size;
}

void init_memfs()
{
  vfs_handler_t *memfs_handler;

  memfs_handler = calloc(sizeof(vfs_handler_t), 1);
  
  memfs_handler->read_func = &read_memfs;
  memfs_handler->write_func = NULL;
  memfs_handler->resolve_func = NULL;
  memfs_handler->creat_func = NULL;
  memfs_handler->unlink_func = NULL;
  memfs_handler->stat_func = NULL;
  
  add_vfs_handler (VFS_MEMFS, memfs_handler);
  
}
