/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

#include <tyros.h>
#include <ntx.h>
#include <core.h>
#include <ntxbase.h>
#include <vfs.h>
#include <filesystems.h>
#include <procfs.h>

id_t resolve_procfs (file_t device, char *path, char **newpath)
{
  int reqproc;
  int prefix;
  char selftest[0x10];
  char *fpath;

  //no more mounts here, please ;)
  *newpath = NULL;
  
  /*
    i haven't got too much time to really MAKE this procFS,
    so this is only a hack to make programs (they depends on 
    sys/proc/self/cmdline & /sys/proc/self/environ) working
  */
  
  strncpy(selftest, path, 4);

  if(!strcmp(selftest, "self")) {
    reqproc = tasks.current;
    fpath = path + 5;
  } else {
    reqproc = atoi(path);

    //let 0 be 'secret system process' ;)
    if(reqproc == 0) {
      return -1;
    }

    fpath = path;
    while(is_digit(*fpath)) {
      fpath++;
    }
    fpath++;
  }
  

  prefix = reqproc << 16;
  
  if(!strcmp(fpath, "environ")) {
    return prefix | PROCFS_ENVIRON;
  } else if(!strcmp(fpath, "cmdline")) {
    return prefix | PROCFS_CMDLINE;
  } else if(!strcmp(fpath, "cwd")) {
    return prefix | PROCFS_CWD;
  } else if(!strcmp(fpath, "exe")) {
    return prefix | PROCFS_EXE;   
  } else if(!strcmp(fpath, "heart")) {
    return prefix | PROCFS_HEART;
  } else {
    return -1;
  }
}

int read_mem (void *membuf, int memsize, int start, int size, void *buffer)
{
  if(start >= memsize) {
    return 0;
  }
  if( (start + size) > memsize) {
    size = memsize - start;
  }

  memcpy(buffer, membuf, size);
  return size;
}

int read_procfs (id_t lid, fsize_t start, fsize_t size, void *buffer, file_t device)
{
  int i;
  char argbuf[0x2000];
  char *p = &argbuf;
  process_t *process;
  int pid;
  int fid;
  int id = (unsigned int) lid;
 
  pid = (id & 0xFFFF0000) >> 16;

  process = pstable->process[pid];
  
  fid = id & 0x0000FFFF;
  
  switch(fid) {
  case PROCFS_ENVIRON:
    memcpyl(argbuf, process->environ, 0x400);
    return read_mem(argbuf, 0x1000, (unsigned int) start,(unsigned int) size, buffer);
    break;

  case PROCFS_CMDLINE:
    memcpyl(argbuf, process->argstr, 0x400);
    return read_mem(argbuf, 0x1000, (unsigned int) start, (unsigned int) size, buffer);
    break;

  case PROCFS_CWD:
    memcpyl(argbuf, process->cwd, 0x400);
    return read_mem(argbuf, strlen(argbuf) + 1, (unsigned int)  start, (unsigned int) size, buffer);
    break;
  case PROCFS_EXE:
    strncpy(argbuf, process->exe, 0x1000);
    return read_mem (argbuf, strlen(argbuf) + 1, (unsigned int)  start, (unsigned int) size, buffer);
    break;
    

  default:
    return -1;
    break;
  }
}

int write_procfs (id_t id, int start, int size, void *buffer, file_t device)
{
  return -1;
}

int unlink_procfs (id_t lid, file_t device)
{
  int pid;
  int fid;
  int id = (unsigned int) lid;
  
  pid = (id & 0xFFFF0000) >> 16;
  
  fid = id & 0x0000FFFF;

  switch(fid) {
  case PROCFS_HEART:
    if(pid > 1) {
      kill_task(pid);
    } else {
      return -1;
    }
    break;
    
  default:
    return -1;
    break;
  }
  
}

void init_procfs()
{
  vfs_handler_t *procfs_handler;

  procfs_handler = calloc(sizeof(vfs_handler_t), 1);
  
  procfs_handler->read_func = &read_procfs;
  procfs_handler->write_func = &write_procfs;
  procfs_handler->resolve_func = &resolve_procfs;
  procfs_handler->unlink_func = &unlink_procfs;
  
  add_vfs_handler (VFS_PROCFS, procfs_handler);
  
}
