/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

#include <tyros.h>
#include <ntx.h>
#include <gui.h>
#include <assert.h>

ntx_windows_table_t *windows;
mouse_state_t *mst;
char *scrcache;

void init_gui_caching()
{
  scrcache = alloc_pmem(mem_to_pages(800*600));
}

void init_gui()
{
  int i;
  int winid;

  //  printf("Initializing NTX GUI...\n");
  
  windows = alloc_pmem(mem_to_pages(sizeof(ntx_windows_table_t)));
  
  for(i = 0; i < WINDOWS_MAX; i++) {
    windows->window[i].owner = -1;
  }
  
  
  mst = alloc_pmem(mem_to_pages(sizeof(mouse_state_t)));

  mst->w = 5;
  mst->h = 5;
  
  gui_mouse_set (100, 100);
  
}

//restore part of screen from cache
int restore_screen (int x, int y, int w, int h)
{
  int i, j;
  char *lfb = hal.lfb;

  for(i = x; i < x+w; i++) {
    for(j = y; j < y+h; j++) {
      lfb[i + j*800] = scrcache[i + j*800];
    }
  }
}

//direct output (no caching)
int putpd (int x, int y, int c)
{
  char *lfb = hal.lfb;
  int off;
  
  off = x + y*800;

  lfb[off] = c;
}

int putp (int x, int y, int c)
{
  char *lfb = hal.lfb;
  int off;

  off = x + y*800;

  lfb[off] = c;
  scrcache[off] = c;
}

int win_mapout (int winid, int x, int y, map_t *map)
{
  return gui_map_out (winid, x, y, map);
}

int win_textout (int winid, int x, int y, char *text)
{
  return gui_text_out (winid, x, y, text);
}

int window (int width, int height)
{
  return gui_create_window(width, height);
}

void draw_framed (int startx, int starty, int width, int height, int pixel)
{
  int i;
  char *lfb = hal.lfb;
  
  for(i = startx; i < startx + width; i++) {
    putpd(i, starty, pixel);
    putpd(i, starty + height, pixel);
  }
  
  for(i = starty; i < starty + height; i++) {
    putpd(startx, i, pixel);
    putpd(startx + width, i, pixel);
  }
  
}


void draw_frame (int startx, int starty, int width, int height, int pixel)
{
  int i;
  char *lfb = hal.lfb;
  
  for(i = startx; i < startx + width; i++) {
    putp(i, starty, pixel);
    putp(i, starty + height, pixel);
  }
  
  for(i = starty; i < starty + height; i++) {
    putp(startx, i, pixel);
    putp(startx + width, i, pixel);
  }
  
}


int gui_display_window (int winid)
{
  int startx, starty;
  int width, height;
  
  width = windows->window[winid].width;
  height = windows->window[winid].height;
  startx = windows->window[winid].startx;
  starty = windows->window[winid].starty;

  draw_frame(startx, starty, width, height, 0x0);

  draw_frame(startx, starty - 20, width, 20, 0x0);

  draw_window_killer (startx + width - 22, starty - 20 + 2);
  
  draw_window_minimizer (startx + 2, starty - 20 + 2);

}

#define MOUSE_SENST 5 //sensitivity

int gui_receive_keypress (int key)
{
  switch(key) {
  case 0x1C: //up
    gui_mouse_move ('U', MOUSE_SENST);
    break;

  case 0x1D: //left
    gui_mouse_move ('L', MOUSE_SENST);
    break;

  case 0x1E: //right
    gui_mouse_move ('R', MOUSE_SENST);
    break;

  case 0x1F: //down
    gui_mouse_move ('D', MOUSE_SENST);
    break;

  default:
    return -1;
    break;
  }

  return 0;
}

int gui_mouse_set (int x, int y)
{
  mst->x = x;
  mst->y = y;
}

int gui_mouse_move (int d, int n)
{
  switch(d) {
  case 'U':
    mst->y-=n;
    break;
  case 'D':
    mst->y+=n;
    break;
  case 'L':
    mst->x-=n;
    break;
  case 'R':
    mst->x+=n;
    break;
  }
  
  draw_mouse_pointer(mst->x, mst->y);
}

int draw_mouse_pointer (int x, int y)
{
  restore_screen (mst->drx, mst->dry, mst->w, mst->h);

  draw_framed (x,y, 3,3,0);
  
  mst->drx = x;
  mst->dry = y;
}



int draw_window_minimizer (int x, int y)
{
  draw_frame (x,y,15,15,0);
}

int draw_window_killer (int x, int y)
{
  draw_frame (x,y,15,15,0);
}

int gui_create_window (int width, int height)
{
  int winid;
  int startx, starty;

  winid = alloc_winid();
  
  if(winid == -1) {
    printf("NTX GUI error: can't alloc window identifier\n");
    return -1;
  }

  startx = 500;
  starty = 200;

  windows->window[winid].owner = tasks.current;
  windows->window[winid].startx = startx;
  windows->window[winid].starty = starty;
  windows->window[winid].width = width;
  windows->window[winid].height = height;
  
  gui_display_window(winid);
  
  return winid;
}

int gui_map_out (int winid,int x, int y, map_t *map)
{
  int winx, winy;
  int i, j;
  char *lfb = hal.lfb;
  int px;

  winx = windows->window[winid].startx + x; 
  winy = windows->window[winid].starty + y;

  printf("gui_map_out, w=%i, h=%i, pix=%X\n", map->width, map->height, map->pix);
  
  for(i = 0; i < map->width; i++) {
    for(j = 0; j < map->height; j++) {
      px = map->pix[j*map->width + i];
      if(px != map->mask) {
	putp(winx + i,winy + j,map->pix[j*map->width + i]);
      }
    }
  }
}

int gui_char_out (int winid, int x, int y, char c)
{
  int winx, winy;

  winx = windows->window[winid].startx;
  winy = windows->window[winid].starty;

  printf("gui_char_out, winx: %i, winy: %i\n", winx, winy);

  display_char (winx + x, winy + y, c);
}

int gui_text_out (int winid, int x, int y, char *text)
{
  int winx, winy, winwidth, winheight;
  int wconwidth, wconheight;
  int gcursor;
  int c;
  int i, j;
  char *lfb = hal.lfb;


  winx = windows->window[winid].startx + 1; 
  winy = windows->window[winid].starty + 1;
  
  winwidth = windows->window[winid].width;
  winheight = windows->window[winid].height;
  
  wconwidth = winwidth/FONT_WIDTH;
  wconheight = winheight/FONT_HEIGHT;
  
  printf("gui_out_text: wconwidth: %i, wconheight: %i\n", wconwidth, wconheight);

  gcursor = 0;
  
  while(*text != 0) {
    
    c = *text;

    switch (c) {
    case '\n': //    
      gcursor+=wconwidth;
      gcursor-=gcursor%wconwidth;
      break;
      
    case 8: //backspace
      gcursor--;
      x = (gcursor%wconwidth) * FONT_WIDTH  + winx;
      y = (gcursor/wconheight) * FONT_HEIGHT + winy;
      display_char(x,y,' ');
    break;
    
    default:
      x = (gcursor%wconwidth) * FONT_WIDTH  + winx;
      y = (gcursor/wconwidth) * FONT_HEIGHT + winy;
      display_char(x,y,c);
      gcursor++;
      break;
  }
    
    
    if(gcursor>=wconwidth*wconheight){
      for(i = winy + FONT_HEIGHT;
	  i <= winy + wconheight*FONT_HEIGHT;
	  i++) {
	
   	
	asm("rep movsl"::"c"(wconwidth*FONT_WIDTH/4),
	    "S" (lfb + i*SCREEN_WIDTH + winx),
	    "D" (lfb + (i - FONT_HEIGHT)*SCREEN_WIDTH + winx));
	
	//	lfb[ (i - 1) * wconwidth * FONT_WIDTH +j ] =
	//	  lfb [i * wconwidth * FONT_WIDTH + j];
	
  
      }
    
      for(i = winy + wconheight*FONT_HEIGHT;
	  i <= winy + wconheight*FONT_HEIGHT + FONT_HEIGHT;
	i++) {
	asm("rep stosl"::"c"(wconwidth*FONT_WIDTH/4),
	    "a" (0x0f0f0f0f),
	    "D" (lfb + (i - FONT_HEIGHT)*SCREEN_WIDTH + winx));
      }
    
      gcursor-=wconwidth;
    }
  
    text++;
  }    
  
restore_screen(winx, winy, winwidth, winheight);  
}

int alloc_winid()
{
  int i;
  
  for(i = 0; i < WINDOWS_MAX; i++) {
    if(windows->window[i].owner == -1) {
      windows->window[i].owner = 0;
      return i;
    }
  }
  
  return -1;
}

int free_winid (int winid)
{
  assert(winid < WINDOWS_MAX);
  windows->window[winid].owner = -1;
}
