/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

#include <tyros.h>
#include "heap.h"

#define PAGE_FREE 0
#define PAGE_ALLOCATED 1

void *heap_grow (heap_t *heap, int npages);

heap_t *heap_alloc ()
{
  return alloc_pmem(mem_to_pages(sizeof(heap_t)));
}

void heap_cleanup(heap_t *heap)
{
  free_pmem(heap->map);
  free_pmem(heap);
}

int heap_init (heap_t *heap)
{
    int i;
    
    heap->size = USER_INITIAL_HEAP_SIZE;

    //    heap->map = (int *) malloc(heap->size*sizeof(int));
    heap->map = alloc_pmem(10);

    for(i = 0; i < heap->size; i++) {
      heap->map[i] = PAGE_FREE;
    }
    
}

int heap_alloc_virtual (heap_t *heap, int npages)
{
  int i, j;
  int freearea;
  int fpage;
  void *allocated;
  unsigned int growth;
  
  freearea = 0;

  for(i = 0; i < heap->size; i++) {
    if (heap->map[i] & PAGE_ALLOCATED) {
      freearea = 0;

    } else {
      freearea++;
      
      if(freearea == npages) {
      
	fpage = i - freearea + 1;
	
	

	heap->map[fpage] = PAGE_ALLOCATED | npages;
	
	for(j = 1; j < npages; j++) {
	  heap->map[fpage + j] = PAGE_ALLOCATED;
	}

	allocated =  USER_HEAP_START + (fpage * 0x1000);
	
	return allocated;
      } else {
	//continue
      }
    }
  }
  
  growth = heap_grow(heap, npages);
  
  allocated =  USER_HEAP_START + growth * 0x1000;
  
  for(i = 0; i < npages; i++) {
    heap->map[growth + i] = PAGE_ALLOCATED;
  }

  return allocated;
}

void *heap_grow (heap_t *heap, int npages) 
{
  int oldsize;
  int new_heap_size;
  int *new_heap_map;
  int i;

  failure("Heap_Grow error\n");
  
  oldsize = heap->size;
  
  new_heap_size = heap->size + npages;
  
  
  //new_heap_map = (int *) realloc(heap->map, new_heap_size*sizeof(int));
  
  for(i = heap->size; i < new_heap_size; i++) {
    heap->map[i] = PAGE_FREE;
  }
  
  
  //heap->map = new_heap_map;
  
  heap->size = new_heap_size;

  return oldsize;
}
