/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

#include <tyros.h>
#include <assert.h>
#include "core.h"

void *it_msg_copy (int task_from, int task_to, void *data, int size);

int is_msg_queue_empty (int tid)
{
  return mailbox_is_empty (pstable->process[tid]->mailbox);
}

int is_msg_queue_full (int tid)
{
  return mailbox_is_full (pstable->process[tid]->mailbox);
}

int ipc_msg_send (int to, message_t *smsg, int flags)
{
  process_t *process;
  mailbox_t *mailbox;
  task_t *task = tasks.task[tasks.current];
  void *newpdata;

  //    printf("Sending message from %i, to %i, ", tasks.current, to);
  
  process = pstable->process[to];
  mailbox = process->mailbox;
  
  isem_disable();

  if(mailbox_is_full(mailbox)) {
    //            printf("%i's mailbox is full, ", to);
    //if the queue is full...
    task->state = TS_SUSPENDED;
    task->reason_why = RS_AWAITING_RECIPIENT;
    task->recipient = to;
    
    isem_enable();
    give_up();

    
  } else {
    isem_enable();
  }

  isem_disable();

  assert(task->state != TS_SUSPENDED);
  
  newpdata = it_msg_copy (tasks.current, to, smsg->pdata, smsg->size);
  
  smsg->pdata = newpdata;
  smsg->from = tasks.current;

  mailbox_put (mailbox, smsg);

  isem_enable();
  
  return 0;
}

int ipc_msg_recv (message_t *smsg, int flags)
{
  message_t *mmsg;
  
  isem_disable();

  //printf("Receiving message (%i),  ", tasks.current);

  

  mmsg = mailbox_get (me->mailbox);
  

  
  smsg->header = mmsg->header;
  smsg->pdata = mmsg->pdata;
  smsg->size = mmsg->size;
  
  isem_enable();
  
  return 0;
}

int ipc_msg_wait ()
{
  task_t *task = tasks.task[tasks.current];

  //    printf("%i is waiting for message..., ", tasks.current);

  task->state = TS_SUSPENDED;
  task->reason_why = RS_AWAITING_MESSAGE;

  give_up();
  assert(task->state != TS_SUSPENDED);

  //  printf("%i is continuing, ", tasks.current);

  return 0;
}

void *it_msg_copy (int task_from, int task_to, void *data, int size)
{
  process_t *to_process, *from_process;
  int from_pagedir, to_pagedir;
  int npages;
  int i, tpte, vpage;

  isem_disable();

  from_process = pstable->process[task_from];
  to_process = pstable->process[task_to];

  from_pagedir = tasks.task[task_from]->pagedir;
  to_pagedir = tasks.task[task_to]->pagedir;
  
  if(from_pagedir == to_pagedir || size == 0) {
    //no need to copy
    return data;
  }

  //  printf("Copied data: %s\n", data);

  npages = (( (unsigned int) (data + size) & 0xFFFFF000) -  ( (unsigned int) data & 0xFFFFF000)) + 1;
  
  //  printf("Copying %i pages of memory from %i to %i\n", npages, task_from, task_to);

  vpage = heap_alloc_virtual (to_process->heap, npages);
  //  printf("Target vpage: %X, npages: %i\n", vpage, npages);


 
  for(i = 0; i < npages; i++) {
    tpte = get_pte ( (unsigned int) data + i*PAGE_SIZE, from_pagedir);
    //    printf("Source PTE: %X\n", tpte);
    //    printf("PTE data: %s\n", (tpte & 0xFFFFF000) + ( (unsigned int)data & 0xFFF));
    set_pte (vpage + i*PAGE_SIZE, to_pagedir, 0x7, tpte);
  }
  
  //  printf("Copy complete\n");

  isem_enable();

  return (void *) (vpage + ( (unsigned int) data & 0xFFF));
}

/* ======== Signals Handling =========== */
/*
  NOTE: only ABORT type signals are supported at present! 
  (no restoration of interrupted program is possible)
*/
int ipc_set_sighandler(signum_t signum, func_t func)
{
  task_t *task = tasks.task[tasks.current];
  printf("Setting sighandler (%i) as %X\n", signum, func);
  task->sighandler[signum] = func;
  return 0;
}
int ipc_send_signal(tid_t tid, signum_t signum)
{
  if (tasks.task[tid] != 0) {
    
    if (tasks.task[tid]->signal == 0) {
      tasks.task[tid]->signal = signum;
      return 0;
    } else {
      return -1;
    }
  } else {
    return -1;
  }
}
