/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

/*
 * ntxbase.c - Neutronix base services:
 * launch(), chdir(), ntx_open(), ntx_close(), ntx_read(), ntx_write()
 */


#include <tyros.h>
#include <ntx.h>
#include <msgs.h>
#include <rpc.h>
#include <ntxbase.h>
#include <core.h>
#include <vfs.h>
#include <filesystems/filesystems.h>
#include <filesystems/infofs.h>
#include <filesystems/devfs.h>

void init_namespace(int nohdd, int nofdd);
void report(char *s);
void display_nsentry(nsentry_t *nsentry);
void display_namespace();

int init_ntxbase()
{
int nohdd = 1, nofdd = 1, displayns = 0;

init_vfs_registry();
init_filesystems();
init_devcache();

init_namespace(nohdd, nofdd);

if(displayns)
 display_namespace();

if(!nohdd)
 mount("/dev/ata00/3/raw", "/mnt/hda", VFS_FAT32);
if(!nofdd)
 mount("/dev/fd0", "/mnt/floppy", VFS_FAT12);
 
mount("/dev/ramdisk", "/mnt/rd", VFS_FAT12);

report("initialization completed\n");
}


void report(char *s)
{
  printf("ntxbase: %s", s);
}


file_t root;

int chdir(char *path)
{
  file_t dir;
  ntx_stat_t statbuf;
  int cwdlen;

  dir = resolve_path(path);
  
  if(dir.id[0]) {
    
    stat_file (dir, &statbuf);
    
    if(statbuf.st_mode & ST_IFDIR) { //is it directory?
      //      printf("chdir to %s succesfull, id = %X\n", path, dir.id[0]);

      if(*path == '/') {
	strcpy(me->cwd, path);
      } else {
	
	
	strcat(me->cwd, path);
      } 

      cwdlen = strlen(me->cwd);
      
      if(me->cwd[cwdlen - 1] != '/') {
	me->cwd[cwdlen] = '/';
	me->cwd[cwdlen + 1] = 0;
      }

      return 0;

    } else {
      return -1;
    }
  
  } else {
    return -1;
  }
}

nsentry_t *add_nsentry (nsentry_t *target, nsentry_t *new, char *name)
{
  int i;
  int newsize;
  char *nname;

  //    i = 0;
  //    while(target->nsentry[i]!=NULL) {
  //      i++;
  //    }
  
  i = target->subcount;
  
  newsize = i + 2;
  
  target->nsentry = realloc(target->nsentry, newsize*sizeof(nsentry_t *));
  target->nsentry[i] = new;
  
  nname = malloc(strlen(name) + 0x8);
  
  strncpy(nname, name, PATH_MAX);
  
  
  target->name = realloc(target->name, newsize*sizeof(char *));
  target->name[i] = nname;
  
  target->nsentry[i + 1] = NULL;
  target->name[i + 1] = NULL;
  
  target->subcount++;
  
  //  printf("%X added to %X as %s (%i)\n", new, target, name, i);
  
  return new;
}


nsentry_t *create_nsentry(int nse_type)
{
  nsentry_t *nse;

  nse = malloc(sizeof(nsentry_t));

  nse->nsentry = malloc(sizeof(nsentry_t *));
  nse->nsentry[0] = NULL;
  nse->name = malloc(sizeof(char *));
  nse->name[0] = NULL;
  
  nse->subcount = 0;
  //default size (for dirs etc)
  nse->nsstatus.size = 0x1000 ;

  nse->nsstatus.vfsid = VFS_VFSROOT;
  nse->nsstatus.ident = nse;

  nse->nse_type = nse_type;
  
  nse->mounted.id[0] = 0;

  return nse;
}


void init_root_directory ()
{
  nsentry_t *root_nsentry;
  
  root_nsentry = create_nsentry(NSE_CONTAINER);
  
  root.id[0] = root_nsentry;

  root.id[1] = -1;
}



void init_namespace(int nohdd, int nofdd)
{
  nsentry_t *root_nsentry;
  nsentry_t *sys, *mnt, *dev, *rd, *info, *devinfo, *kernel, *proc, *devproc; 
  nsentry_t *control, *devcontrol;
  file_t file, rdfile, infofile, procfile, controlfile;
  char buffer[0x10000];

  init_root_directory();
  root_nsentry = root.id[0];


  sys = add_nsentry(root_nsentry,
		    create_nsentry(NSE_CONTAINER),
		    "sys");
  
  mnt = add_nsentry(root_nsentry,
		    create_nsentry(NSE_CONTAINER),
		    "mnt");
  
  dev = add_nsentry(root_nsentry,
		    create_nsentry(NSE_CONTAINER),
		    "dev");
  
  kernel = add_nsentry(dev,
		       create_nsentry(NSE_CONTAINER),
		       "ntx");
  
  devinfo = add_nsentry(kernel,
			create_nsentry(NSE_OBJECT),
			"info");
  
  devproc = add_nsentry(kernel,
			create_nsentry(NSE_OBJECT),
			"proc");

  devcontrol = add_nsentry(kernel,
			   create_nsentry(NSE_OBJECT),
			   "control");
  
  info = add_nsentry(sys,
		     create_nsentry(NSE_CONTAINER),
		     "info");
  
  proc = add_nsentry(sys,
		     create_nsentry(NSE_CONTAINER),
		     "proc");
  control = add_nsentry(sys,
			create_nsentry(NSE_CONTAINER),
			"control");




  rd = add_nsentry(mnt, create_nsentry(NSE_CONTAINER), "rd");
  if(!nohdd)
   add_nsentry(mnt, create_nsentry(NSE_CONTAINER), "hda");
  if(!nofdd)
   add_nsentry(mnt, create_nsentry(NSE_CONTAINER), "floppy");

  init_devices(dev, nohdd, nofdd);

  
  info->nsstatus.vfsid = VFS_INFOFS;
  
  infofile.id[0] = devinfo;
  infofile.vfsid[0] = 0;
  infofile.id[1] = -1;
  
  info->mounted = infofile;

  control->nsstatus.vfsid = VFS_CONTROLFS;
  controlfile.id[0] = devcontrol;
  controlfile.vfsid[0] = 0;
  controlfile.id[1] = -1;
  
  control->mounted = controlfile;

  proc->nsstatus.vfsid = VFS_PROCFS;
  procfile.id[0] = devproc;
  procfile.vfsid[0] = 0;
  procfile.id[1] = -1;
  
  proc->mounted = procfile;

  
}

int stat_file (file_t file, ntx_stat_t *buffer)
{
  int dvid;
  int (*stat_func)(id_t, void*, file_t);
  nsentry_t *device;
  file_t devfile;

  dvid = 0;
  while(file.id[dvid] != -1) {
    dvid++;
  }
  
  if(dvid == 1) {
    device = file.id[0];
    
    stat_vfs (device, buffer, file);
    
    return 0;

  } else if(dvid == 2) {
    stat_func = vfs_registry->vfs_handler[file.vfsid[0]]->stat_func;
    
    devfile.id[0] = file.id[0];
    devfile.vfsid[0] = file.vfsid[0];
    devfile.id[1] = -1;

    stat_func(file.id[1], buffer, devfile);
    
    return 0;

  } else {
    //printf("stat_file error: dvid > 2\n");
    return -1;
  }
}

id_t creat_file (char *filename, int mode)
{
  int i, fnlen;
  char *changed = NULL;
  file_t dir;
  char cwdbuff[0x1000];
  char *dirname;
  char *filetitle;
  id_t result;

  if(strchr(filename, '/') == NULL) {
    strcpy(cwdbuff,me->cwd);
    dirname = cwdbuff;
    filetitle = filename;
  } else {
    fnlen = strlen(filename);
    for(i = fnlen; i >= 0; i--) {
      if(filename[i] == '/') {
	filename[i] = 0;
	changed = &filename[i];
	filetitle = &filename[i + 1];
	break;
      } 
    }
    dirname = filename;
  }

  dir = resolve_path(dirname);
  if(dir.id[0] == 0 || dir.id[0] == -1) {
    //    printf("'%s': no such directory\n", dirname);
    if(changed != NULL) {
      *changed = '/';
    }
    return -1;
  }
  
  i = 0;

  do {
    i++;
  } while (dir.id[i] != -1 && dir.id[i] != 0);
  
  //can't create files on a first level filesystem (VFSroot)
  if(i < 2) {
    if(changed != NULL) {
      *changed = '/';
    }
    return -1;
  }

  result = vfs_registry->vfs_handler[dir.vfsid[i - 2]]->creat_func (filetitle, mode, dir);
  
  if(changed != NULL) {
    *changed = '/';
  }
  
  return result;
}

int write_file (file_t file, fsize_t start, fsize_t fsize, void *buffer)
{
    int dvid;
  int (*write_func)(id_t, fsize_t, fsize_t, void*, file_t);
  nsentry_t *device;
  file_t devfile;


  dvid = 0;
  while(file.id[dvid] != -1) {
    //    printf("file.id[%i] = %X\n", dvid, file.id[dvid]);
    dvid++;
  }
  
  if(dvid == 1) {
    device = file.id[0];
    
    write_func = vfs_registry->vfs_handler[device->nsstatus.vfsid]->write_func;

    return write_func (device->nsstatus.ident, start, fsize, buffer, file);

  } else if(dvid == 2) {
    write_func = vfs_registry->vfs_handler[file.vfsid[0]]->write_func;
    
    devfile.id[0] = file.id[0];
    devfile.vfsid[0] = file.vfsid[0];
    devfile.id[1] = -1;
    return write_func(file.id[1], start, fsize, buffer, devfile);

  } else {
    //printf("write_file error: dvid > 2\n");
    return -1;
  }
  

} 

int unlink_file (file_t file)
{
 int dvid;
  int (*unlink_func)(id_t, file_t);
  nsentry_t *device;
  file_t devfile;


  dvid = 0;
  while(file.id[dvid] != -1) {
    //    printf("file.id[%i] = %X\n", dvid, file.id[dvid]);
    dvid++;
  }
  
  if(dvid == 1) {
    device = file.id[0];
    
    unlink_func = vfs_registry->vfs_handler[device->nsstatus.vfsid]->unlink_func;
    
    return unlink_func (device->nsstatus.ident, file);

  } else if(dvid == 2) {
    unlink_func = vfs_registry->vfs_handler[file.vfsid[0]]->unlink_func;
    
    devfile.id[0] = file.id[0];
    devfile.vfsid[0] = file.vfsid[0];
    devfile.id[1] = -1;
    return unlink_func(file.id[1], devfile);

  } else {
    //printf("unlink_file error: dvid > 2\n");
    return -1;
  }
  

}

int mkdir (file_t file, char *name, int mode)
{
  int (*mkdir_func) (id_t, char *, int, file_t);
  file_t device;
  nsentry_t *devns;
  id_t dirid;
  int vfsid;
  
//  unpack_directory(file, &device, &dirid);
  
  devns = file.id[0];
  
  vfsid = devns->nsstatus.vfsid;
  
  //  printf("VFSID: %i\n", vfsid);
  
  mkdir_func = vfs_registry->vfs_handler[vfsid]->mkdir_func;
  
  return mkdir_func(dirid, name, mode, device);
}

int read_file (file_t file, fsize_t start, fsize_t fsize, void *buffer)
{
  int dvid;
  int (*read_func)(id_t, fsize_t, fsize_t, void*, file_t);
  nsentry_t *device;
  file_t devfile;


  dvid = 0;
  while(file.id[dvid] != -1) {
    //    printf("file.id[%i] = %X\n", dvid, file.id[dvid]);
    dvid++;
  }
  
  if(dvid == 1) {
    device = file.id[0];
    
    read_func = vfs_registry->vfs_handler[device->nsstatus.vfsid]->read_func;
    
    return read_func (device->nsstatus.ident, start, fsize, buffer, file);
    
  } else if(dvid == 2) {
    read_func = vfs_registry->vfs_handler[file.vfsid[0]]->read_func;
    
    devfile.id[0] = file.id[0];
    devfile.vfsid[0] = file.vfsid[0];
    devfile.id[1] = -1;
    return read_func(file.id[1], start, fsize, buffer, devfile);

  } else {
    //printf("read_file error: dvid > 2\n");
    return -1;
  }
  

} 



int recdepth;

void display_namespace()
{
  recdepth = 0;

  printf("/");
  display_nsentry(root.id[0]);
}


void display_nsentry(nsentry_t *nsentry)
{
  int i, j;
  
  
  i = 0;
  
  recdepth++;
  
  puts("\n");
  

  while(nsentry->nsentry[i] != NULL) {
    
    if(strcmp(nsentry->name[i], ".") && strcmp(nsentry->name[i], "..")) {
    
      for(j = 0; j < recdepth; j++) {
	puts("   ");
      }
      
      puts("/");
      puts(nsentry->name[i]);
      
      
      display_nsentry(nsentry->nsentry[i]);
    }
    
    i++;
  }

  recdepth--;
}






char **environ;

char *search_path (char *file, char *envpath, char *newpath)
{
  char *PATH[0x100];
  char buffer[PATH_MAX];
  char *start;
  int argc;
  int i;
  int finish;
  char *ret = NULL;;
  file_t f;
  
  strcpy(buffer, envpath);

  
  
  argc = 0;
  i = 0;
  finish = 0;

  start = buffer;
  for(;;) {
    if(buffer[i] == ';' || buffer[i] == 0) {
      
      if(buffer[i] != 0) {
	buffer[i] = 0;
	PATH[argc] = start;
	argc++;
	

	start = &buffer[i + 1];
      } else {
	PATH[argc] = start;
	argc++;
	start = &buffer[i + 1];
	finish = 1;
      }      

      if(*start == 0 || finish == 1) {
	break;
      }
    }
    
    i++;
  }

  for(i = 0; i < argc; i++) {
    
   
    strcpy(newpath, PATH[i]);
    strcat(newpath, "/");
    strcat(newpath, file);

    //printf("trying %s (argc == %i)\n", newpath, argc);
    f = resolve_path(newpath);
    
    if(f.id[0] != NULL) {
      //yeah! we've found it :)
      
      return 1;
    }
  } 
  
  
  
  return 0;
}



int launch(char *path, char *argv[], char *envp[])
{
  nsentry_t *nsentry;
  char *envpath;
  void *buffer;
  fsize_t size;
  int tid;
  int i;
  char spath[PATH_MAX];
  file_t backdir;
  file_t file;
  ntx_stat_t statbuf;
  int sres;
  int reg;
  /*
  if(*path != '/') {
    environ = envp;
    envpath = getenv("PATH");
    
    sres = search_path (path, envpath, spath);
    
    if(sres != NULL) {
      path = spath;
    }
    
  }
  */
  
  //  printf ("Launching %s (cwd: %s)\n", path, me->cwd);

  file = resolve_path(path);


  if(file.id[0] != NULL) {
    
    stat_file(file, &statbuf);
    
    //if it is directory
    if(statbuf.st_mode & ST_IFDIR) {
      printf("launch: '%s' is a directory!\n", path);
      return -1;
    }

    size = statbuf.st_size;

    //printf("file size: %i\n", size);
    
    buffer = alloc_pmem(mem_to_pages(size));
    
    
    read_file (file, 0, size, buffer);
    
    isem_disable();
    
    
    tid =  load_process(buffer, size, 0);

    if(tid != -1) {
    
      if(argv != NULL) {
	copy_argv(tid, tasks.current, argv,  pstable->process[tid]->argstr);
      }
    
    
      if(envp != NULL) {
	//Yes! Luckily internal structure of ARGV and ENVP are same
	copy_argv(tid, tasks.current, envp,  pstable->process[tid]->environ);
      }
      
      strcpy(pstable->process[tid]->exe, path);
      
    }
    
    free_pmem(buffer);
    
    isem_enable();
    
    return tid;
    
  } else {
    return -1;
  }


}

//klibc wrappers
int open(char *path, int flags)
{
  return sys_ntx_open(path, flags);
}

int read (int fd, void *buff, int len)
{
  return sys_ntx_read(fd, buff, len);
}

int close(int fd)
{
  return sys_ntx_close(fd);
}

int unlink(int fd)
{
  return sys_ntx_unlink(fd);
}


int sys_ntx_write (int fd, void *buff, int len)
{
  int res;

  verify_read (buff, len);

  if(fd >= USER_FDS_MAX) {
    return -1;//invalid descriptor
  }
  if(me->fd[fd].file.id[0] == 0) {
    return -1;//invalid descriptor
  }

  res = write_file (me->fd[fd].file, me->fd[fd].fpos, len, buff);
  
  me->fd[fd].fpos += res;

  return res;
}

int sys_ntx_read (int fd, void *buff, int len)
{
  int res;

  verify_write(buff, len);

  if(fd >= USER_FDS_MAX) {
    return -1;//invalid descriptor
  }
  if(me->fd[fd].file.id[0] == 0) {
    return -1;//invalid descriptor
  }
  
  res = read_file (me->fd[fd].file, me->fd[fd].fpos, len, buff);

  me->fd[fd].fpos += res;

  //  printf("Read operation fpos + %i = %i\n", res, me->fd[fd].fpos);
  
  return res;
}

int sys_ntx_mount (char *device, char *dir, char *fstype, int general_flags, char *fsflags)
{
  int vfsid;
  
  vfsid = get_fs_by_name(fstype);
  
  if(vfsid != -1) {
    return mount(device, dir, vfsid);
  } else {
    return -1;
  }
}

int sys_ntx_stat (int fd, ntx_stat_t *statbuf)
{
  verify_write(statbuf, sizeof(ntx_stat_t));

  if(fd >= USER_FDS_MAX) {
    return -1;//invalid descriptor
  }
  if(me->fd[fd].file.id[0] == 0) {
    return -1;//invalid descriptor
  }
  
  return stat_file(me->fd[fd].file, statbuf);
}

int sys_ntx_close (int fd)
{
  if(fd >= USER_FDS_MAX) {
    return -1;//invalid descriptor
  }
  if(me->fd[fd].file.id[0] == 0) {
    return -1;//invalid descriptor
  }
  
  me->fd[fd].file.id[0] = 0;

  return 0;
}

int sys_ntx_unlink (char *filename)
{
  file_t file;
  int fd;

  file = resolve_path (filename);
  
  if(file.id[0] == NULL) {
    return -1;//no such file
  } else {
    return unlink_file(file);
  }
}

int sys_ntx_mkdir (char *name, int mode)
{
  char *d;
  char *dirname;
  file_t dir;
  int result;

  dirname = alloc_pmem(mem_to_pages(PATH_MAX));

  d = strrchr(name, '/');
  
  if(d != NULL) {
    strncpy(dirname, name, (d - name));
  } else {
    strncpy(dirname, "");
    d = name - 1;
  }
  
  //  printf("resolving dir '%s'\n", dirname);
  
  dir = resolve_path (dirname);
  
  if(dir.id[0] == NULL) {
    result = -1;
  } else {
    result = mkdir (dir, d + 1, mode);
  }

  free_pmem(dirname);

  return result;
}

int sys_ntx_open (char *filename, int flags)
{
  file_t file;
  int fd;

  file = resolve_path (filename);
  
  if(file.id[0] == NULL) {

    if( !(flags & O_CREAT)) { 

      return -1;//no such file
    
    } else {
      creat_file (filename, 0);
      
      file = resolve_path (filename);

      if(file.id[0] == NULL) {
	return -1;
      }
    }
    
  }
  fd = process_alloc_fd (me);
  if(fd == -1) {
    return -1;//no free descriptors left
  } else {
    me->fd[fd].file = file;
    
    //we set initial file position to zero
    /*
      TODO: check flags such as O_APPEND
    */
    me->fd[fd].fpos = 0;
    
    
    return fd;
  }
  
  
}


int create_file (char *path)
{
  
}

