/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */
#include <tyros.h>
#include <vfs.h>
#include "core.h"
#include "process.h"
#include "rpc.h"
#include "ntxbase.h"
#include "exec/exec.h"

process_t *process_alloc()
{
  return alloc_pmem(mem_to_pages(sizeof(process_t)));
}

void process_cleanup(process_t *process)
{
  //  mailbox_cleanup(process->mailbox);
  heap_cleanup(process->heap);
  free_pmem(process);
}


int process_init (process_t *process)
{
  int i;

  process->rpc_count = 0;
  //  process->rpc_handler = malloc(1);

  //  printf("Process init...");
  process->heap = heap_alloc();
  heap_init (process->heap);
  //  printf("Heap OK. ");
  
  //  process->mailbox = mailbox_alloc();
  //  mailbox_init (process->mailbox);
  //  printf("Mailbox OK. ");

  if(tasks.current != 0) {
    process->nsentry = me->nsentry;
    strcpy(process->cwd, me->cwd);
  } else {
    process->nsentry = root;
    strcpy(process->cwd, "/");
  }

  for(i = 0; i < USER_FDS_MAX; i++) {
    process->fd[i].file.id[0] = 0;
  }

  strcpy(process->argstr, "\0");
}

int process_alloc_fd (process_t *process)
{
  int i;
  
  for(i = 0; i < USER_FDS_MAX; i++) {
    if(process->fd[i].file.id[0] == 0) {
      return i;
    }
  }

  return -1;
}

int process_free_fd (process_t *process, int fdid)
{
  if(process->fd[fdid].file.id[0] != 0) {
    process->fd[fdid].file.id[0] = 0;
    return 0;
  } else { //invalid FD
    return -1;
  }
}

int process_add_rpc_handler (process_t *process, int vector, void *handler, int flags)
{
  int newsize;

  if(process->rpc_count <= vector) {
    newsize = vector * sizeof(void*);
    
    process->rpc_count = newsize + 1;

    process->rpc_handler = realloc(process->rpc_handler, newsize);
  }    

  process->rpc_handler[vector] = handler;
  
  return 0;
}

int process_invoke_rpc_handler (int tid, int vector, message_t *msg, int flags)
{
  int (*handler)(message_t *msg);
  int result;
  process_t *process;
  message_t *newmsg;
  int backtid;

  process = pstable->process[tid];
  
  if(vector >= process->rpc_count) {
    return -1;
  } else {
    
    switch (flags) {
      
    case RPC_REALTIME:
      
      //newmsg = malloc(sizeof(message_t));
      
      //memcpy (newmsg, msg, sizeof(message_t));
      
      
      handler = process->rpc_handler[vector];

      isem_disable();

      backtid = tasks.current;

      switch_task (tid);
      

      //     printf("NEWMSG: %X\n", newmsg);
      
      result = handler(msg);
      
      switch_task (backtid);

      //free(newmsg);

      isem_enable();
      
      return result;

    default:
      return -1;
      break;
    }
  }
  
}


task_image_t load_image (process_t *process, void *image, int size)
{
  task_image_t nullimage;

  if( * (int *) image == 0x464C457F) {
    return load_elf_image(process, image, size);
  } else {
    panic("load_image: Unknown executable format!\n");
  }
}

int process_load (process_t *process, void *image, int size, int pl)
{
  int i;
  void * ptask_entry;
  unsigned int pagedir;
  task_image_t task_image;
  unsigned int total_pages;
  unsigned int phys_page;

  //  ptask_entry = (void *)  alloc_pmem(mem_to_pages(size));


  pagedir = build_page_directory();
  
  process->pagedir = pagedir;
  
  memcpyl(pagedir, hal.pagedir, 512); //copy first half of the system pagedir

  task_image = load_image (process, image, size);
  
  
  //one PDE (4 megabytes) is allocated for stack
  set_pde(USER_STACK_START, pagedir, VM_PAGE_ALLOCATED);
  
  

  return add_task(task_image, pagedir, pl);
}

char *copy_argv (int task_to, int task_from, char **argv, char *argstr)
{

  process_t *to_process, *from_process;
  int from_pagedir, to_pagedir;
  int i;
  int total_len;

  char buffer[0x1000];

  from_process = pstable->process[task_from];
  to_process = pstable->process[task_to];

  from_pagedir = tasks.task[task_from]->pagedir;
  to_pagedir = tasks.task[task_to]->pagedir;
  
  i = 0;
  total_len = 0;
  



  do{
    //    printf("reallocating (new %i bytes)\n", total_len + strlen(argv[i]));
    
    strcpy( argstr + total_len, argv[i]);
    
    total_len += (strlen(argv[i]));

    *(argstr + total_len + 1) = 0;
    
    
    
    total_len++;
    i++;

  } while(argv[i] != NULL);


  
  *(argstr + total_len + 2) = 0;



  return argstr;
}
