/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

#include <tyros.h>
#include <ntxbase.h>
#include "syscalls.h"

void **syscall_handler;
int syscalls_count;


int install_syscall(int vector, void *handler)
{
  int newsize;

  if(vector >= syscalls_count) {
    syscall_handler = realloc(syscall_handler, (vector + 1)*sizeof(void *));
    syscalls_count = vector + 1;
  } 
  
  syscall_handler[vector] = handler;
}

int syscall_open(int ebx, int ecx, int edx, int esi, int edi)
{
  return sys_ntx_open (ebx, ecx);
}

int syscall_close (int ebx, int ecx, int edx, int esi, int edi)
{
  return sys_ntx_close (ebx);
}

int syscall_mount (int ebx, int ecx, int edx, int esi, int edi)
{
  return sys_ntx_mount (ebx, ecx, edx, esi, edi);
}

int syscall_read (int ebx, int ecx, int edx, int esi, int edi)
{
  return sys_ntx_read (ebx, ecx, edx);
}

int syscall_unlink (int ebx, int ecx, int edx, int esi, int edi)
{
  return sys_ntx_unlink (ebx);
}

int syscall_mkdir (int ebx, int ecx, int edx, int esi, int edi)
{
  return sys_ntx_mkdir (ebx, ecx);
}

int syscall_write (int ebx, int ecx, int edx, int esi, int edi)
{
  return sys_ntx_write (ebx, ecx, edx);
}

int syscall_fork (int ebx, int ecx, int edx, int esi, int edi)
{
  return process_fork();
}

int syscall_load_thread(int ebx, int ecx, int edx, int esi, int edi)
{
  return load_thread(ebx);
}

int syscall_ipc_msg_send (int ebx, int ecx, int edx, int esi, int edi)
{
  return ipc_msg_send(ebx, ecx, edx);
}

int syscall_ipc_msg_recv (int ebx, int ecx, int edx, int esi, int edi)
{
  return ipc_msg_recv (ebx, ecx);
}

int syscall_ipc_msg_wait (int ebx, int ecx, int edx, int esi, int edi)
{
  return ipc_msg_wait();
}

int syscall_morecore(int ebx, int ecx, int edx, int esi, int edi)
{
  return morecore(ebx);
}


int syscall_exit (int ebx, int ecx, int edx, int esi, int edi)
{
  return exit(ebx);
}

int syscall_execve (int ebx, int ecx, int edx, int esi, int edi)
{
  return execve(ebx, ecx, edx);
}

int syscall_waitpid (int ebx, int ecx, int edx, int esi, int edi)
{
  return waitpid(ebx, ecx, edx);
}

int syscall_launch (int ebx, int ecx, int edx, int esi, int edi)
{
  return launch(ebx, ecx, edx);
}

int syscall_chdir (int ebx, int ecx, int edx, int esi, int edi)
{
  return chdir(ebx);
}

int syscall_stat (int ebx, int ecx, int edx, int esi, int edi)
{
  return sys_ntx_stat(ebx, ecx);
}

int syscall_window (int ebx, int ecx, int edx, int esi, int edi)
{
  return window(ebx, ecx);
}

int syscall_win_textout (int ebx, int ecx, int edx, int esi, int edi)
{
  return win_textout (ebx, ecx, edx, esi);
}

int syscall_win_mapout (int ebx, int ecx, int edx, int esi, int edi)
{
  return win_mapout (ebx, ecx, edx, esi);
}

void init_syscalls()
{
  syscall_handler = malloc(1*sizeof(void *));

  install_syscall(SC_LOAD_THREAD, &syscall_load_thread);
  install_syscall(SC_IPC_MSG_SEND, &syscall_ipc_msg_send);
  install_syscall(SC_IPC_MSG_RECV, &syscall_ipc_msg_recv);
  install_syscall(SC_IPC_MSG_WAIT, &syscall_ipc_msg_wait);
  install_syscall(SC_MORECORE, &syscall_morecore);

  install_syscall(SC_OPEN, &syscall_open);
  install_syscall(SC_CLOSE, &syscall_close);
  install_syscall(SC_READ, &syscall_read);
  install_syscall(SC_WRITE, &syscall_write);

  install_syscall(SC_FORK, &syscall_fork);

  install_syscall(SC_EXIT, &syscall_exit);
  install_syscall(SC_EXECVE, &syscall_execve);
  install_syscall(SC_WAITPID, &syscall_waitpid);

  install_syscall(SC_LAUNCH, &syscall_launch);

  install_syscall(SC_CHDIR, &syscall_chdir);

  install_syscall(SC_STAT, &syscall_stat);

  install_syscall(SC_UNLINK, &syscall_unlink);
  install_syscall(SC_MKDIR, &syscall_mkdir);
  install_syscall(SC_MOUNT, &syscall_mount);
  
  install_syscall(SC_WINDOW, &syscall_window);
  install_syscall(SC_WIN_TEXTOUT, &syscall_win_textout);
  install_syscall(SC_WIN_MAPOUT, &syscall_win_mapout);
}




unsigned int syscall (unsigned int eax,
		      unsigned int ebx,
		      unsigned int ecx,
		      unsigned int edx,
		      unsigned int esi, 
		      unsigned int edi,
		      unsigned int ext)
{
  int err;
  int (*handler)(int, int, int, int, int);
  int ret;
  int esp;
  extern int syscall_esp;
  extern tss_t *sys_tss;

  esp = syscall_esp;
  
  asm("sti");
  
  
  
  if(eax < syscalls_count) {
    handler = syscall_handler[eax];
    
    ret =  handler(ebx, ecx, edx, esi, edi);


  } else {
    
    ret = -1;
  }

  asm("cli");
  
  //interprivelege int pushes 5 dwords onto stack (cs, eip, eflags, ss, esp)
  sys_tss->esp0 = esp + 5*4; 
  return ret;
}
