/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

#include <tyros.h>
#include <assert.h>
#include "core.h"
#include "process.h"
#include "ntxbase.h"
#include "../hal/ia32/include/vm.h"

void violation(void *area, int size);

int verify_write (void *area, unsigned int size)
{
  int i;
  int npages;
  int pte;

  npages = mem_to_pages(size);
  
  for(i = 0; i < npages; i++) {
    pte = get_pte (area + i*PAGE_SIZE, me->pagedir);
    
    if(pte == -1) {
      violation(area, size);
    }
    
    if( (pte & 7) == 7) {
      //everything ok
    } else {
      
      if(pte != VM_PAGE_ALLOCATED) {
	violation(area, size);
      }

    }

  }
}

int verify_read (void *area, unsigned int size)
{
  int i;
  int npages;
  int pte;

  npages = mem_to_pages(size);
  
  for(i = 0; i < npages; i++) {
    pte = get_pte (area + i*PAGE_SIZE, me->pagedir);
    
    if(pte == -1) {
      violation(area, size);
    }
    
    if( (pte & 5) == 5) {
      //everything ok
    } else {
      if(pte != VM_PAGE_ALLOCATED) {
	violation(area, size);
      }
    }

  }
}



void violation(void *area, int size)
{
  printf("Access violation when accessing %X \n", area);
  kill_task(tasks.current);
}

void failure (char *s)
{
  printf("%zcritical failure on task %d: %sTerminating...%z\n", 
					COLOR_LRED, tasks.current, s, 7);
  kill_task(tasks.current);
}
