/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

#ifndef _VFS_H_
#define _VFS_H_

#include <tyros.h>
#include <ntx.h>
#include <ntxbase.h>

#define VFS_ROOT 0

typedef struct {
 file_t file;
 fsize_t fpos;
} fd_t;

typedef struct {
  int (*read_func)(id_t, fsize_t, fsize_t, void*, file_t);
  int (*write_func)(id_t, fsize_t, fsize_t, void*, file_t);
  int (*stat_func) (id_t, void *, file_t);
  int (*mount_func)(file_t, file_t);
  int (*mkdir_func)(id_t, char *, int, file_t);
  id_t (*resolve_func) (file_t, char *, char **);
  int (*unlink_func) (id_t, file_t);
  id_t (*creat_func) (char *,int, file_t);
  
  char *name;
} vfs_handler_t;


typedef struct {
  int nhandlers;
  vfs_handler_t **vfs_handler;
} vfs_registry_t;

extern vfs_registry_t *vfs_registry;

extern file_t root;
extern file_t current_nsentry;

file_t resolve_path(char *path);

#endif
