/*
 *   handlers.c - HAL-level interrupt handlers 
 *
 *
 *   Copyright (c) 2003 Lonesome & OSDev Group
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */  


#include <tyros.h>



EXCEPTION_HANDLER(divide_error)
{
  failure("Divide error\n");
  //  printf("%zDivide Error\n", COLOR_LRED);
  //  asm("cli");
  //  for(;;);
}

EXCEPTION_HANDLER(nmi_interrupt)
{
  failure("NMI Interrupt\n");
  //  printf("%zNMI Interrupt\n", COLOR_LRED);
  //  asm("cli");
  //  for(;;);
}

EXCEPTION_HANDLER(debug_exception)
{
  failure("Debug exception\n");
  //  printf("%zDebug Exception\n", COLOR_LRED);
  //  asm("cli");
  //  for(;;);
}

EXCEPTION_HANDLER(breakpoint_trap)
{
  failure("Breakpoint trap\n");
  //  printf("%zBreakpoint\n", COLOR_LRED);
  //  asm("cli");
  //  for(;;);
}

EXCEPTION_HANDLER(overflow_exception)
{
  failure("Overflow exception\n");
  //  printf("%zOverflow Exception\n", COLOR_LRED);
  //  asm("cli");
  //  for(;;);
}

EXCEPTION_HANDLER(bound_exception)
{
  failure("Bound range exceeded exception\n");
  //  printf("%zBound Range Exceeded Exception\n", COLOR_LRED);
  //  asm("cli");
  //  for(;;);
}

EXCEPTION_HANDLER(invalid_opcode)
{
  //printf("%zInvalid Opcode Exception\n", COLOR_LRED);
  failure("Invalid opcode exception\n");
  //  asm("cli");
  //  for(;;);
}

EXCEPTION_HANDLER(no_math)
{
  failure("device not available exception\n");
  //  printf("%zDevice Not Available Exception\n", COLOR_LRED);
  //  asm("cli");
  //  for(;;);
}

EXCEPTION_HANDLER(double_fault)
{
  failure("double fault\n");
  //  printf("%zDouble Fault\n", COLOR_LRED);
  //  asm("cli");
  //  for(;;);
}

EXCEPTION_HANDLER(invalid_tss)
{
  failure("invalid tss exception\n");
  //  printf("%zInvalid TSS Exception\n", COLOR_LRED);
  //  asm("cli");
  //  for(;;);
}

EXCEPTION_HANDLER(segment_not_present)
{
  failure("segment not present exception\n");
  //  printf("%zSegment Not Present\n", COLOR_LRED);
  //  asm("cli");
  //  for(;;);
}

EXCEPTION_HANDLER(stack_fault)
{
  failure("stack fault exception\n");
  //  printf("%zStack Fault Exception\n", COLOR_LRED);
  //  asm("cli");
  //  for(;;);
}

EXCEPTION_HANDLER(general_protection)
{
  failure("general protection fault\n");
  //  panic("General Protection Fault. EIP: %X\n",  address);
  //  asm("cli");
  //  for(;;);
}

EXCEPTION_HANDLER(page_fault)
{
  unsigned int cr2;

  //we MUST move CR2 to stack variable immediately,
  //cause another page fault can occur
  asm("movl %%cr2, %0":"=d"(cr2));
  
#ifdef VM_DEBUG
  printf("Page Fault: CR2: %X\n", cr2);
#endif
  
  if(vm_pagefault(cr2) == VM_UNRESOLVED) {

    
    
  printf("Invalid Page Fault Exception (%X) when accessing %X\nFaulting task: %i\nEIP: %X\n",
	errcode, cr2, tasks.current, address);
  dump(address, 0x30);

  kill_task(tasks.current);
  for(;;);
  }
  
#ifdef VM_DEBUG
  printf("%zPage fault resolved by VMM, continuing...\n", COLOR_LGREEN);
#endif 

}

EXCEPTION_HANDLER(fpu_error)
{
  failure("FPU error\n");
  //  printf("%zx87 FPU Floating Point Error\n", COLOR_LRED);
  //  asm("cli");
  //  for(;;);
}

EXCEPTION_HANDLER(alignment_exception)
{
  failure("alignment check exception\n");
  //  printf("%zAlignment Check Exception\n", COLOR_LRED);
  //  asm("cli");
  //  for(;;);
}

EXCEPTION_HANDLER(machine_exception)
{
  failure("machine exception\n");
  //  printf("%zMachine Check Exception\n", COLOR_LRED);
  //  asm("cli");
  //  for(;;);
}

EXCEPTION_HANDLER(simd_error)
{
  failure("SIMD exception\n");
  //  printf("%zSIMD Floating Point Exception\n", COLOR_LRED);
  //  asm("cli");
  //  for(;;);
}

//syscall handler (int 0x77)

int syscall_return;
int syscall_ebp;
int syscall_esp;

int syscall_cs;
int syscall_eflags;
int syscall_ss;
int syscall_esp;

asm(
    "\n  .globl int_syscall"
    "\n  int_syscall:"
    "\n  cli"
    //    "\n  popl syscall_return"
    //    "\n  popl syscall_cs"
    //    "\n  popl syscall_eflags"
    //    "\n  popl syscall_esp"
    //    "\n  popl syscall_ss"

    //    "\n  movl %ebp, syscall_ebp"
    
    "\n movl %esp, syscall_esp"

    //    "\n  pushl syscall_ss"
    //    "\n  pushl syscall_esp"
    //    "\n  pushl syscall_eflags"
    //    "\n  pushl syscall_cs"
    //    "\n  pushl syscall_return"
    "\n  pusha"

    "\n  pushl $0"
    "\n  pushl %edi"
    "\n  pushl %esi"
    "\n  pushl %edx"
    "\n  pushl %ecx"
    "\n  pushl %ebx"
    "\n  pushl %eax"

    "\n  call syscall"
    "\n  addl $28, %esp"

    "\n  movl %eax, sc_result"
    "\n  popa"
    "\n  movl sc_result, %eax"
    "\n  iretl"
    "\n  sc_result: .long 0"
    );


void receive_scancode(int);

IRQ_HANDLER(irq_keyboard)
{
  unsigned char cr;

  asm("cli");

  cr = inportb(0x61);

  outportb(0x61, cr | 0x80);
  
  receive_scancode(inportb(0x60));
  
  outportb(0x61, cr);
  asm("sti");
}
