/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

#ifndef _TYPES_H_
#define _TYPES_H_



typedef struct hal_st {
  unsigned int mode;

  unsigned int memory_amount;
  unsigned int gdt_addr;
  unsigned int pagedir;
  unsigned int nameserver_binary;
  unsigned int nameserver_size;
  
  unsigned int ramdisk_binary;
  unsigned int ramdisk_size;

  unsigned int console_drv_binary;
  unsigned int console_drv_size;
  unsigned int ramdisk_drv_binary;
  unsigned int ramdisk_drv_size;

  void (*putchar)(char);
  
  void *elf_file;
  
  unsigned int lfb;
} hal_t;

typedef unsigned int func_t;
typedef unsigned int signum_t;
typedef int tid_t;
typedef unsigned int port_t;
typedef unsigned int gate_t;
typedef unsigned int uint;
typedef unsigned int udword;

#define NULL 0
#define TRUE (1 == 1)
#define FALSE (1 == 0)

typedef struct tss_st {
  unsigned int backlink;
  unsigned int esp0;
  unsigned int ss0;
  unsigned int esp1;
  unsigned int ss1;
  unsigned int esp2;
  unsigned int ss2;
  unsigned int pdbr; //cr3
  unsigned int eip;
  unsigned int eflags;
  unsigned int eax;
  unsigned int ecx;
  unsigned int edx;
  unsigned int ebx;
  unsigned int esp;
  unsigned int ebp;
  unsigned int esi;
  unsigned int edi;
  unsigned int es;
  unsigned int cs;
  unsigned int ss;
  unsigned int ds;
  unsigned int fs;
  unsigned int gs;
  unsigned int ldtr;
  unsigned short tflags;
  unsigned short bitmap;
} __attribute__((packed)) tss_t;


typedef struct {
  int header;
  void *pdata;
  int size;
  int from;  
} message_t;


typedef struct stack0_st {
  
  udword eax;
  udword ebx;
  udword ecx;
  udword edx;
  udword esi;
  udword edi;
  udword ebp;
  
  udword gs;
  udword fs;
  udword es;
  udword ds;
  
  udword eip;
  udword cs;
  udword eflags;
  udword esp;
  udword ss;
} __attribute__((packed)) stack0_t;

typedef struct {
  int start;
  int size;
  void *entry;
} task_image_t;


typedef struct task_st {
  int selector;
  func_t sighandler[0x10];
  int signal;
  
  int pagedir;

  int state;
  int reason_why;
  int waiting_port;
  int recipient;
  int recip_port;

  int msg_flag;

  int msgn;

  int pl;
  int cs;
  
  unsigned int priority;
  
  task_image_t image;
  stack0_t *stack0;
  void *stack0init;
  tss_t *tss;
} task_t;

struct tasks_table_st {
  int count;
  int current;
  int current_qid;
  int queue[1024];
  task_t **task;
};




#endif
