/*
 *   ints.c - HAL-level interrupt handling facilities 
 *
 *
 *   Copyright (c) 2003 Lonesome & OSDev Group
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */  

#include <tyros.h>

int isem;

void init_isem()
{
  isem = 0;
}

void int_handler(int vector);

void dummy_handler(void);
asm(
    "\n  dummy_handler:"
    "\n  movb $0x20, %al"
    "\n  outb %al, $0x20"
    "\n  iretl"
    );

void irq_keyboard(void);

void init_pic()
{
  /*
    set base vector of 1st controller to 0x20,
    of 2nd controller to 0x28
  */
  
  outportb(0x20, 0x15);
  outportb(0xA0, 0x15);
  
  outportb(0x21, 0x20);
  outportb(0xA1, 0x28);
  
  outportb(0x21, 0x04);
  outportb(0xA1, 0x02);
  
  outportb(0x21, 0x0D);
  outportb(0xA1, 0x09);
  
  
}

void init_interrupts()
{
  
  
  install_auto_handlers();
  
  
  i_install(0, &divide_error, 0x8e);
  i_install(1, &debug_exception, 0x8e);
  i_install(2, &nmi_interrupt, 0x8e);
  i_install(3, &breakpoint_trap, 0x8e);
  i_install(4, &overflow_exception, 0x8e);
  i_install(5, &bound_exception, 0x8e);
  i_install(6, &invalid_opcode, 0x8e);
  i_install(7, &no_math, 0x8e);
  i_install(8, &double_fault, 0x8e);
  i_install(10, &invalid_tss, 0x8e);
  i_install(11, &segment_not_present, 0x8e);
  i_install(12, &stack_fault, 0x8e);
  i_install(13, &general_protection, 0x8e);
  i_install(14, &page_fault, 0x8e);
  i_install(16, &fpu_error, 0x8e);
  i_install(17, &alignment_exception, 0x8e);
  i_install(18, &machine_exception, 0x8e);
  i_install(19, &simd_error, 0x8e);

  i_install(0x77, &int_syscall, 0xef);

  i_install(0x21, &irq_keyboard, RING0 | INTERRUPT_GATE);
  
  i_setup();
}


void i_install(unsigned char vector, void (*func)(), unsigned char type) 
{
  char * idt_table= (char *) IDT_TABLE;
  
  unsigned char i;
  unsigned char b[8];

  b[0]=  (unsigned int)func & 0x000000FF;
  b[1]=( (unsigned int)func & 0x0000FF00) >> 8;
  b[2]=SYS_CODE_SELECTOR;
  b[3]=0;
  b[4]=0;
  b[5]=type;
  b[6]=( (unsigned int)func & 0x00FF0000) >> 16;
  b[7]=( (unsigned int)func & 0xFF000000) >> 24;

  
  for(i=0;i<8;i++){
    *(idt_table+vector*8+i)=b[i];
  }
  
  
}


void i_setup()
{
  unsigned short *table_limit = (unsigned short *) IDT_REG;
  unsigned int *table_address = (unsigned int *) (IDT_REG+2); 

  *table_limit = 256*8 - 1;
  *table_address = IDT_TABLE;
  asm("lidt 0(,%0,)"::"a"(IDT_REG));
}


void isem_enable()
{
  
  if(isem <= 1) {
    isem = 0;
    
    asm("sti");
  } else {
    isem--;
  }
  
}

void isem_disable()
{
  asm("cli");
  isem++;
}

void i_enable()
{
  asm("sti");
}

void i_disable()
{
  asm("cli");
}
