/*
 *   main.c - HAL initialization routines 
 *
 *
 *   Copyright (c) 2003 Lonesome & OSDev Group
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */  

#include <tyros.h>
#include <gui.h>
#include <assert.h>

void init_mm(void);
void init_taskman(void);
void init_interrupts(void);
void init_hal(void);
void display_char(int x, int y, char c);

hal_t hal;


char gdt_reg[6];

char cfont[256][FONT_WIDTH * FONT_HEIGHT];



asm("sys_font:");
asm(".incbin \"char.mt\"");

extern int sys_font;

void init_hal()
{
  unsigned int *gdt_addr = (unsigned int *) (gdt_reg + 2);

  asm("sgdt gdt_reg");
  
  hal.gdt_addr = *gdt_addr;
  
  hal.ramdisk_binary = 0x400000;
}

void self_test()
{
  assert(sizeof(task_t) < PAGE_SIZE);
}


/*
  According to multiboot specification, bootloader
  MUST provide kernel with working segment registers,
  but NEED NOT provide it with working GDT
*/

asm(
    "\n  .align 4"
    "\n  gdt:"

    /* NULL descriptor */
    "\n  .word 0, 0, 0, 0"
   
    /* CODE segment, 
       DPL == 0
       BASE == 0
       LIMIT == 4gb
    */
    "\n  .byte 0xFF"
    "\n  .byte 0xFF"
    "\n  .byte 0x00"
    "\n  .byte 0x00"
    "\n  .byte 0x00"
    "\n  .byte 0x9A"
    "\n  .byte 0xCF"
    "\n  .byte 0x00"


    /* DATA segment
       DPL == 0
       BASE == 0
       LIMIT == 4gb
    */
    "\n  .byte 0xFF"
    "\n  .byte 0xFF"
    "\n  .byte 0x00"
    "\n  .byte 0x00"
    "\n  .byte 0x00"
    "\n  .byte 0x92"
    "\n  .byte 0xCF"
    "\n  .byte 0x00"

    /* CODE segment
       DPL == 3
       BASE == 0
       LIMIT == 4gb
    */
    "\n  .byte 0xFF"
    "\n  .byte 0xFF"
    "\n  .byte 0x00"
    "\n  .byte 0x00"
    "\n  .byte 0x00"
    "\n  .byte 0xFA"
    "\n  .byte 0xCF"
    "\n  .byte 0x00"
    
    /* DATA segment
       DPL == 3
       BASE == 0
       LIMIT == 4gb
    */
    "\n  .byte 0xFF"
    "\n  .byte 0xFF"
    "\n  .byte 0x00"
    "\n  .byte 0x00"
    "\n  .byte 0x00"
    "\n  .byte 0xF2"
    "\n  .byte 0xCF"
    "\n  .byte 0x00"
    
    
    "\n  .word 0, 0, 0, 0"

    "\n  .byte 0xFF"
    "\n  .byte 0xFF"
    "\n  .byte 0, 0, 0"
    "\n  .byte 0x9A"
    "\n  .byte 0, 0"

    "\n  .byte 0xFF, 0xFF"
    "\n  .byte 0, 0, 0"
    "\n  .byte 0x92"
    "\n  .byte 0, 0"
    
    "\n  .fill 256, 8, 0"

    "\n  gd_reg:"
    "\n  .word 256*8 - 1"
    "\n  .long gdt"
    
    );

extern int gdt;

/*
void setup_gdt()
{
  void *gdp = &gdt;

  printf("New GDT: %X\n", gdp);

  asm("lgdt gd_reg");
  
  init_hal();
  printf("GDT: %X\n", hal.gdt_addr);
  dump(hal.gdt_addr, 0x30);

  asm("movw $0x10, %ax \n movw %ax, %ds\n");
}
*/

extern struct {
  unsigned int 	 width;
  unsigned int 	 height;
  unsigned int 	 bytes_per_pixel; /* 3:RGB, 4:RGBA */ 
  unsigned char	 pixel_data[562 * 143 * 3 + 1];
} gimp_image;

int gcursor;
char gattrib;
char gbackgr;

void gputchar(int c)
{
  int x, y;
  int i, j;
  char *lfb = hal.lfb;

  switch (c) {
  case '\n': //    
    gcursor+=CON_WIDTH;
    gcursor-=gcursor%CON_WIDTH;
    break;
    
  case 8: //backspace
    gcursor--;
    x = (gcursor%CON_WIDTH) * FONT_WIDTH  + CON_BASEX;
    y = (gcursor/CON_WIDTH) * FONT_HEIGHT + CON_BASEY;
    display_char(x,y,' ');
    break;
    
  default:
    x = (gcursor%CON_WIDTH) * FONT_WIDTH  + CON_BASEX;
    y = (gcursor/CON_WIDTH) * FONT_HEIGHT + CON_BASEY;
    display_char(x,y,c);
    gcursor++;
    break;
  }

     
  if(gcursor>=CON_WIDTH*CON_HEIGHT){
    for(i = CON_BASEY + FONT_HEIGHT;
	i <= CON_BASEY + CON_HEIGHT*FONT_HEIGHT;
	i++) {
      
   	
	asm("rep movsl"::"c"(CON_WIDTH*FONT_WIDTH/4),
	    "S" (lfb + i*SCREEN_WIDTH + CON_BASEX),
	    "D" (lfb + (i - FONT_HEIGHT)*SCREEN_WIDTH + CON_BASEX));
	
	//	lfb[ (i - 1) * CON_WIDTH * FONT_WIDTH +j ] =
	//	  lfb [i * CON_WIDTH * FONT_WIDTH + j];
	
  
    }
    
    for(i = CON_BASEY + CON_HEIGHT*FONT_HEIGHT;
	i <= CON_BASEY + CON_HEIGHT*FONT_HEIGHT + FONT_HEIGHT;
	i++) {
      asm("rep stosl"::"c"(CON_WIDTH*FONT_WIDTH/4),
	  "a" (0x0f0f0f0f),
	  "D" (lfb + (i - FONT_HEIGHT)*SCREEN_WIDTH + CON_BASEX));
    }
    

    /*
    for(i = CON_BASEY + (CON_HEIGHT + 1) *FONT_HEIGHT; 
	i <= CON_BASEY + (CON_HEIGHT + 2) *FONT_HEIGHT;
	i++) {
    asm("rep stosl"::"c"(CON_WIDTH*FONT_WIDTH/4),
	"D" (lfb + i*SCREEN_WIDTH + CON_BASEX),
	"a"(0x0f0f0f0f));
	}*/
    gcursor-=CON_WIDTH;
  }
  
      
  
}

void display_char(int x, int y, char c)
{
  int i,j;
  char *lfb = hal.lfb;

  for(i = 0; i < FONT_WIDTH; i++) {
    for(j = 0; j < FONT_HEIGHT; j++) {
      if(cfont[c][j*FONT_WIDTH + i] == 0x6f) {
	putp(x + i, y + j, gattrib);
      } else {
	putp(x + i, y + j, gbackgr);
      }
    }
  }

}

extern void (*tputchar)(char);
extern int edata;

typedef struct boot_data_struct {
  int lfb;
  int mode;
  int load_from;
  int elf_file;
} boot_data_t;

extern char *scrcache;

void hal_main(boot_data_t *bootdata)
{
  void (*ctor_entry)();
  int i, j;
  char *lfb;
  int width, height;
  int target;
  int byte1, byte2, byte3;
  int eodv = &edata;


  /* performing all necessary initializations */
  hal.lfb = bootdata->lfb;

  hal.mode = bootdata->mode;

  hal.elf_file = bootdata->elf_file;

  init_pic();


  if(bootdata->mode == 0) {
    hal.putchar = &tputchar;
  } else {

    
    lfb = hal.lfb;
    
    asm(
	"\n  movl $0x0f0f0f0f, %%eax"
	"\n  movl $800*600/4, %%ecx"
	"\n  rep  stosl"
	::"D"(hal.lfb)
	);
   
     
    
    
    memcpy(cfont, &sys_font, 256*FONT_WIDTH*FONT_HEIGHT);
    
    
    
    gcursor = 0;
    gattrib = 0x9;
    gbackgr = 0xf;
    
    
    
    hal.putchar = &gputchar;
  }  
  
  
  
  init_isem();
  isem_disable();
  
  init_hal();
  
  
  init_tty();


  //  setup_gdt();
  
  clear();




  
  self_test();
  
  init_mm();  
  
  if(hal.mode != 0) {
    init_gui_caching();
    
    asm(
	"\n  movl $0x0f0f0f0f, %%eax"
	"\n  movl $800*600/4, %%ecx"
	"\n  rep  stosl"
	::"D"((int)scrcache)
	);

    draw_frame (CON_BASEX - 1, CON_BASEY - 1, 
		CON_WIDTH*FONT_WIDTH + 2, CON_HEIGHT*FONT_HEIGHT + 2, 0x0d);
    
  }  
  
  printf("The Neutronix\n");
  printf("Compiled %s, %s\n", __DATE__, __TIME__);
  



  init_paging();

  init_vm();

  init_interrupts();

  init_keyboard();

  init_taskman();


  
  core_entry();
  
  
  for(;;){
    give_up();
  }
}

void restart_system()
{
void (*reboot)() = 0;
reboot = (void*) 0xf000fff0;
}

void dump(char *addr, int num)
{
  int i;
  puthexi(addr);
  puts(": ");
  for(i=1;i<=num;i++){
    puthex(*(addr+i-1));   
    putchar(' ');
    if(i%16 == 0){
      putchar('\n');
    
      if(i!=num){
	puthexi(addr+i);
	puts(": ");
      }    

    }
  }
}
