/*
 *   paging.c - paging control facilities 
 *
 *
 *   Copyright (c) 2003 Lonesome & OSDev Group
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */


#include <tyros.h>

#define DEFAULT_PAGE_DIR_ENTRY 7

void enable_paging();

void init_paging()
{
  unsigned int pagedir;

  pagedir = build_page_directory();
  
  //map_memory_page(0x12345678, 0x80000000, pagedir);

  identity_map(pagedir, 0x3);
  

  load_cr3(pagedir);
  enable_paging();
  
  hal.pagedir = pagedir;

}

/*
unsigned int create_user_address_space(unsigned int physical_start, unsigned int size)
{
  unsigned int user_page_dir;
  unsigned int basepage, userstack;
  unsigned int i;

  user_page_dir = build_page_directory();
  
  basepage = physical_start;
  
  //  printf("Building system area\n");

  memcpyl(user_page_dir, hal.pagedir, 0x200);

  //  identity_map(user_page_dir, 0x3);
  
  //  printf("Image size: %i pages\n", mem_to_pages(size));

  for(i = 0; i < mem_to_pages(size); i++) {
    map_memory_page(USER_ADDRESS_SPACE_START + i*0x1000, basepage + i*0x1000, user_page_dir, 0x7);
  }

  //  printf("Building stack area (%i pages)\n", USER_SPACE_FOR_STACK/PAGE_SIZE);
  //  map_memory_page(USER_STACK_START, userstack + 0x1000 - 4, user_page_dir,  0x7);
  
  set_pde(USER_STACK_START, user_page_dir, VM_PAGE_ALLOCATED);
  
  //  printf("COMPLETE\n");

  return user_page_dir;
}
*/


void identity_map(unsigned int *pagedir, unsigned int pflags)
{
  int i;
  int memory_pages;
  int lfbpages;

  lfbpages = (800*600)/0x1000;

  memory_pages = hal.memory_amount/0x1000;

  for(i = 0; i < memory_pages; i++) {
    
    map_memory_page(i * 0x1000, i * 0x1000, pagedir, pflags);
  }

  if(hal.lfb != 0) {
    for(i = 0; i < lfbpages; i++) {
      map_memory_page(i*0x1000 + LFB_MAPPING, i * 0x1000 + hal.lfb, pagedir, pflags);
    }
  }
  
}

unsigned int build_page_directory()
{
  unsigned int * pagedir;
  int i;
  
  pagedir = alloc_pmem(1);

  for (i = 0; i < 1024; i++) {
    pagedir[i] = 0;
  }
  
  return pagedir;
}

int copy_pagedir(int new_pagedir_addr, int old_pagedir_addr)
{
  int i;
  unsigned int *new_pagedir, *old_pagedir;
  unsigned int table_addr;
  unsigned int new_table_addr;
  unsigned int table_flags;
  unsigned int pde, pte;
  
  new_pagedir = new_pagedir_addr;
  old_pagedir = old_pagedir_addr;
  
  for(i = 0; i < 1024; i++) {
    
    //determine PDE
    pde = old_pagedir[i];
    
    if(pde & 1) {
      //if page table is present
      
      
      //determine page table address & flags
      table_addr = pde & 0xFFFFF000;
      table_flags = pde & 0x00000FFF;
      
      new_table_addr = task_alloc_pmem(tasks.current, 1);
      
      copy_pagetable(new_table_addr, table_addr);
      
      new_pagedir[i] = new_table_addr | table_flags;
      //      printf("Copying PDE, new: %X, old: %X\n", new_pagedir[i], old_pagedir[i]);
      

    } else {
      //if page table is not present
      new_pagedir[i] = old_pagedir[i];
    }
    
    
  }
}

int copy_pagetable (unsigned int new_table_addr, unsigned int old_table_addr)
{
  int i;
  unsigned int *new_pagetable, *old_pagetable;
  unsigned int pte;

  new_pagetable = new_table_addr;
  old_pagetable = old_table_addr;
  
  for(i = 0; i < 1024; i++) {
    
    pte = old_pagetable[i];

    if(pte & 1) {
      //page present
      new_pagetable[i] = (old_pagetable[i] & 0xFFFFFFFD) | VM_PAGE_COPY_ON_WRITE;
      old_pagetable[i] = (old_pagetable[i] & 0xFFFFFFFD) | VM_PAGE_COPY_ON_WRITE;
      
    } else {
      //page isn't present
      
      new_pagetable[i] = old_pagetable[i];
      
    }
    
  }
  
}

unsigned int get_pte (unsigned int from, unsigned int pagedir_addr)
{
  unsigned int page_dir_index, page_table_index;
  unsigned int pagetable, page_table_entry;
  unsigned int *page_table_addr;
  unsigned int *pagedir = (unsigned int *) pagedir_addr;
  
  //high 10 bits of page address are index in the page directory
  page_dir_index = from >> 22;
  
  //bits 12-21 are index in the page table
  page_table_index = (from >> 12) & 0x3ff;
  
  //low 12 bits are out of our interest
  //(those are offset in the page)
  
  page_table_entry = (unsigned int *) pagedir [page_dir_index];
  
  pagetable = page_table_entry >> 12;
  
  
  if(pagetable == 0) {
    
    return 0;
    
  } else {
    page_table_addr = (pagetable << 12);
  }
  
  return page_table_addr[page_table_index]; 
  
}

void set_pde (unsigned int from, unsigned int pagedir_addr, unsigned int value)
{
  unsigned int page_dir_index, page_table_index;
  unsigned int pagetable, page_table_entry;
  unsigned int *page_table_addr;
  unsigned int *pagedir = (unsigned int *) pagedir_addr;
  
  //high 10 bits of page address are index in the page directory
  page_dir_index = from >> 22;
  
  //bits 12-21 are index in the page table
  page_table_index = (from >> 12) & 0x3ff;
  
  pagedir [page_dir_index] = value;
}

unsigned int get_pde (unsigned int from, unsigned int pagedir_addr)
{
  unsigned int page_dir_index, page_table_index;
  unsigned int pagetable, page_table_entry;
  unsigned int *page_table_addr;
  unsigned int *pagedir = (unsigned int *) pagedir_addr;
  
  //high 10 bits of page address are index in the page directory
  page_dir_index = from >> 22;
  
  //bits 12-21 are index in the page table
  page_table_index = (from >> 12) & 0x3ff;
  
  return pagedir [page_dir_index];
}

void set_pte (unsigned int from, unsigned int pagedir_addr, unsigned int pde_flags, unsigned int value)
{
  unsigned int page_dir_index, page_table_index;
  unsigned int pagetable, page_table_entry;
  unsigned int *page_table_addr;
  unsigned int *pagedir = (unsigned int *) pagedir_addr;
  
  //high 10 bits of page address are index in the page directory
  page_dir_index = from >> 22;
  
  //bits 12-21 are index in the page table
  page_table_index = (from >> 12) & 0x3ff;
  
  //low 12 bits are out of our interest
  //(those are offset in the page)
  
  
  page_table_entry = (unsigned int *) pagedir [page_dir_index];
  
  pagetable = page_table_entry >> 12;
  
  
  
  if(pagetable == 0) {
    page_table_addr = task_alloc_pmem(tasks.current, 1);
    memsetl(page_table_addr, 0, 1024);
    pagedir[page_dir_index] = (unsigned int)  page_table_addr | 0x7;
  } else {
    page_table_addr = (pagetable << 12);
  }
  
  
  
  page_table_addr[page_table_index] = value;
  
}


/* Initialize virtual memory for demand loading 
   (set it to VM_PAGE_ALLOCATED) */
void init_for_demand (void *addr, int npages, int pagedir)
{
  int i;
  void *p;

  for(i = 0; i < npages; i++) {
    p = addr + i*PAGE_SIZE;
    set_pte (p, pagedir, 0x7, VM_PAGE_ALLOCATED);
  }
  
}

/* map virtual address FROM to physical address TO */
void map_memory_page(unsigned int from, unsigned int to, unsigned int pagedir_addr, int pflags)
{
  set_pte (from, pagedir_addr, pflags, to | pflags);
}

/* reload pagedir */
void load_cr3(unsigned int pagedir)
{
  //  printf("Loading CR3 with %X...\n", pagedir);
  asm("movl %0, %%cr3"::"d"(pagedir));
}

void enable_paging()
{
  printf("Enabling paging....");

  asm("\n  movl %cr0, %eax"
      "\n  orl $0x80000000, %eax"
      "\n  movl %eax, %cr0");

  hal.lfb = LFB_MAPPING;

  printf("OK.\n");
}
