/*   
 *
 *   taskman.c - HAL-level task manager 
 *
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

#include <tyros.h>
#include <assert.h>
#include <ktty.h>

void int_timer(void);
int is_ready_for_run(int);
void task_cleanup(task_t *task);
void add_used_page (int tid, unsigned int page);
void load_task_register(int selector);
void add_tss_descriptor(unsigned int selector, unsigned int base, unsigned char type);
void init_tids_allocator(void);
unsigned int alloc_tid(void);
void free_tid(unsigned int tid);

tss_t *sys_tss;

struct tasks_table_st tasks;

unsigned int ticks_counter = 0;


int oldstack;
int old_stack_bottom;

void give_up()
{
  asm("int $32");
}


//check if tasks.task[TID] is ready to be executed
//(tasks.task[TID]->state == TS_READY)

int terminal_owner;

int action_on_ctrlc(int vcid)
{
  printf("\nTerminated\n");
  kill_task(terminal_owner);
}

int is_ready_for_run (int tid)
{
  task_t *task;
  int task_ready;
  int task_to_run;
  
  tasks.current = tid;

  task = tasks.task[tasks.current];

  task_ready = 0;
  
  if(task->state == TS_READY) {
    task_ready = 1;
    
  } else if (task->state == TS_SUSPENDED) {
    
	//	printf("%i is suspended\n", tasks.current);

	if (task->reason_why == RS_AWAITING_MESSAGE) {

	  //printf("%i is waiting for message...\n", task_to_run);
	  if(!is_msg_queue_empty(tasks.current)) {
	    task->state = TS_READY;
	    task->reason_why = RS_NONE;
	  }
	  
	} else if (task->reason_why == RS_AWAITING_RECIPIENT) {
	  
	  if(!is_msg_queue_full(task->recipient)) {
	    task->state = TS_READY;
	    task->reason_why = RS_NONE;
	  }
	  
	} else if (task->reason_why == RS_WAITING_PID_DIE) {
	  
	  //	  	  move(0, 0);
		  //	  	  printf("%i waits until %i dies - %i\n", tasks.current, task->recipient, ticks_counter);
	  
	  if(tasks.task[task->recipient] == NULL) {
	    task->state = TS_READY;
	    task->reason_why = RS_NONE;
	    //	    printf("%i is dead, continuing...\n", task->recipient);
	    task_ready = 1;
	  }
	} else if (task->reason_why == RS_WAITING_WHILE_ZERO) {
	  
	  if( * (int *) task->recipient != 0) {

	    task->state = TS_READY;
	    task->reason_why = RS_NONE;
	    
	    task_ready = 1;
	  }
	  
	}
	
      }
      

  if(task_ready == 1) {
    return 1;
  } else {
    return 0;
  }
  
}


#define PRIORITY_MAX 0x10000
#define PRIORITY_MIN 0x10

//this func knows what to switch, but doesn't know how
int scheduler()
{
  int task_to_run, task_ready;
  task_t *task;
  int qid_to_run;
  int qid;
  int nc;
  int i;
  unsigned int maxp, maxt, maxq; //maxPriority, maxTID, maxQID

  nc = 0;

  if(tasks.count != 0) {
    
    qid = tasks.current_qid;

    /*
  
    //simple round-robin scheduler:
    //walks through the queue of tasks and schedules it to run if it's ready
    
    do {
      
      if( (qid + 1) >= tasks.count) {
	qid = 0;
      } else {
	qid++;
      }
      
      nc++;
      
      if(nc >= tasks.count) {
	qid = 0;
	break;
      }

    } while (!is_ready_for_run(tasks.queue[qid]));
    
    */
    
    maxp = 0;
    maxq = 0;

    for(i = 0; i < tasks.count; i++) {
      task = tasks.task[tasks.queue[i]];
      
      if( (task->priority > maxp) && is_ready_for_run(tasks.queue[i])) {
	maxp = task->priority;
	maxq = i;
      }
      
    }
    
    //printf("chosen task: %i (priority %i, total %i)\n", tasks.queue[maxq], maxp, tasks.count);
    
    for(i = 0; i < tasks.count; i++) {
      task = tasks.task[tasks.queue[i]];
      
      if(i == maxq) {
	//chosen task
	if(task->priority > (PRIORITY_MIN + tasks.count -2) ) {
	  task->priority -= (tasks.count - 2);
	}
      } else {
	if(i != 0) {
	  if(task->priority < PRIORITY_MAX) {
	    task->priority++;
	  }
	}
      }
    }
    
    qid = maxq;

    tasks.current = tasks.queue[qid];
    tasks.current_qid = qid;
    
    
    
    //    printf("%i (qid: %i) is ready for run\n", tasks.queue[qid], qid);

    //CORE-layer has its own current task pointer (process_t *me)
    //tell core to switch it
    switch_core_task();
    
    return tasks.queue[qid];

  } else {

    panic("SCHEDULER PANIC: NO TASKS TO RUN!\n");
    return -1;
  }


}

int old_cs;


int lastsaved;
int lastrestored;

//this func knows how to switch, but doesn't know what
stack0_t *switch_stack(void)
{
  int task_to_run;
  task_t *before_switch_task, *after_switch_task;
  
  before_switch_task = tasks.task[tasks.current];

  
  /*
  if(old_cs == SYS_CODE_SELECTOR && (tasks.current != 0)
     && before_switch_task->state != TS_SUSPENDED) {
    return oldstack;
  }
  */
  
  before_switch_task->stack0 = oldstack;
  

  before_switch_task->cs = old_cs;
  
  
  
  task_to_run = scheduler();
  /*
  tasks.current = 1;
  task_to_run = 1;
  switch_core_task();
  */

  after_switch_task = tasks.task[task_to_run];

  //reload cr3 with new pagedir if it's not the same
  if(after_switch_task->pagedir != before_switch_task->pagedir) {
    //  printf("Reloading CR3 with %X\n", tasks.task[task_to_run]->pagedir);
    //  printf("Switching to %i\n", task_to_run);
  
    //  display_stack0(tasks.task[task_to_run]->stack0);
 
    load_cr3(after_switch_task->pagedir);
  } else {
    
    //some self-testing
    if(after_switch_task->stack0 == 0) {
      printf("Switching to %i: INCORRECT STACK!!!\n", task_to_run);
      for(;;);
    }
    
  }


  
  if(after_switch_task->cs == USER_CODE_SELECTOR) {
    sys_tss->esp0 = ((void *) after_switch_task->stack0) + 16*4;
  } else {
    sys_tss->esp0 = ((void *) after_switch_task->stack0) + 14*4;
  }
  return (after_switch_task->stack0);
}


/*
  software task-switch: 

  1) push all regs to stack
  2) save esp to oldstack (we'll patch SYS_TSS later)
  3 call switch_stack()
  4) pop all regs from new stack
  5) iret
  
*/

void int_78(void);

asm(
    "\n  int_78:"
    
    "\n  cli"
    "\n  incl ticks_counter"

    // save old CS
    //(we'll need it later to determine what ring we were in)
    "\n  popl old_eip"
    "\n  popl old_cs"
    "\n  pushl old_cs"
    "\n  pushl old_eip"
    

    "\n  pushl   %ds"
    "\n  pushl   %es"
    "\n  pushl   %fs"
    "\n  pushl   %gs"
    "\n  pushl   %ebp"
    "\n  pushl   %edi"
    "\n  pushl   %esi"
    "\n  pushl   %edx"
    "\n  pushl   %ecx"
    "\n  pushl   %ebx"
    "\n  pushl   %eax"

    "\n  movl %esp, %eax"
    
    //    "\n  call switch_stack"
    "\n  movl %eax, %esp"
    
    "\n  movb $0x20, %al\n"
    "\n  outb %al, $0x20\n"
    
    
    "\n  popl    %eax"
    "\n  popl    %ebx"
    "\n  popl    %ecx"
    "\n  popl    %edx"
    "\n  popl    %esi"
    "\n  popl    %edi"
    "\n  popl    %ebp"

    "\n  popl    %gs"
    "\n  popl    %fs"
    "\n  popl    %es"
    "\n  popl    %ds"
    
    "\n  sti"
    
    "\n  iret"
);

asm(
    "\n  old_eflags: .long 0"
    "\n  old_eip: .long 0"
    "\n  .globl int_timer"
    "\n  int_timer:"
    
    "\n  cli"
    "\n  incl ticks_counter"

    // save old CS
    //(we'll need it later to determine what ring we were in)
    "\n  popl old_eip"
    "\n  popl old_cs"
    "\n  pushl old_cs"
    "\n  pushl old_eip"
    

    "\n  pushl   %ds"
    "\n  pushl   %es"
    "\n  pushl   %fs"
    "\n  pushl   %gs"
    "\n  pushl   %ebp"
    "\n  pushl   %edi"
    "\n  pushl   %esi"
    "\n  pushl   %edx"
    "\n  pushl   %ecx"
    "\n  pushl   %ebx"
    "\n  pushl   %eax"

    "\n  movl %esp, oldstack"
    
    "\n  call c_int_timer\n"
    
    "\n  call switch_stack"
    "\n  movl %eax, %esp"
    
    
    "\n  movb $0x20, %al\n"
    "\n  outb %al, $0x20\n"
    
    
    "\n  popl    %eax"
    "\n  popl    %ebx"
    "\n  popl    %ecx"
    "\n  popl    %edx"
    "\n  popl    %esi"
    "\n  popl    %edi"
    "\n  popl    %ebp"

    "\n  popl    %gs"
    "\n  popl    %fs"
    "\n  popl    %es"
    "\n  popl    %ds"
    
    "\n  iret"
    );



int usyscall(int eax, int ebx, int ecx, int edx, int esi, int edi)
{
  int retvalue;
  asm("int $0x77":"=a"(retvalue):"a"(eax), "b"(ebx), "c"(ecx), "d"(edx),"S"(esi),"D"(edi));
  return retvalue;
}




void display_stack0(stack0_t *stack0)
{
  printf("CS: %X   EIP: %X   EFLAGS: %X\n"
	 "DS: %X, ES: %X, FS: %X, GS: %X, SS: %X", 
	 stack0->cs, stack0->eip, stack0->eflags,
	 stack0->ds, stack0->es, stack0->fs, stack0->gs, stack0->ss);
}

void display_tss(tss_t *tss)
{
  
  printf("Backlink: %X\n"
	 "ESP0: %X, SS0: %X, ESP1: %X, SS1: %X\n"
	 "ESP2: %X, SS2: %X\n"
	 "CR3 : %X\n"
	 "EIP : %X, EFLAGS: %X\n"
	 "EAX : %X, EBX: %X, ECX: %X, EDX: %X\n"
	 "ESI : %X, EDI: %X, ESP: %X, EBP: %X\n"
	 "CS  : %X, DS : %X, ES : %X, FS : %X\n"
	 "GS  : %X, SS : %X\n"
	 "LDTR: %X, IO Bitmap: %X\n",
	 tss->backlink,
	 tss->esp0, tss->ss0, tss->esp1, tss->ss1, tss->esp2, tss->ss2,
	 tss->pdbr,
	 tss->eip, tss->eflags,
	 tss->eax, tss->ebx, tss->ecx, tss->edx, tss->esi, tss->edi, tss->esp, tss->ebp,
	 tss->cs, tss->ds, tss->es, tss->fs, tss->gs, tss->ss,
	 tss->ldtr, tss->bitmap);
}



void kill_task(int tid)
{
  int i;
  unsigned int qid;
  task_t *task;
  
  asm("cli");
  
  qid = -1;
  
  //  printf("Killing TID %i (tasks.current: %i)\n", tid, tasks.current);

  for(i = 0; i < tasks.count; i++) {
    if(tasks.queue[i] == tid) {
      qid = i;
    }
  }

  assert(qid != -1);
  
  for(i = qid; i < tasks.count; i++) {
    tasks.queue[i] = tasks.queue[i+1];
  }
  
  tasks.count--;
  
  task = tasks.task[tid];

  tasks.task[tid] = NULL;
  
  kill_core_task(tid);
  task_cleanup(task);


  //  printf("%zTask %i (QID: %i) killed\n", COLOR_LGREEN, tid, qid);
  
  //  printf("Queue looks like: \n");
  /*for(i = 0; i < tasks.count; i++) {
    printf("tasks.queue[%i] == %i\n", i, tasks.queue[i]);
    }*/

  asm("sti");
  
  if(tasks.current == tid) {
    //we don't wanna continue execution if it was suicide
    give_up();
  }
  
}

void free_page_tables(int *pagedir)
{
  int i, j;
  int *table;;

  for(i = 512; i < 1024; i++) {
    if (pagedir[i] & 1) {
      table = pagedir[i] & 0xFFFFF000;
      for(j = 0; j < 1024; j++) {
	if(table[j] & 1) {
	  //	  printf("Freeing page: %X\n", table[j] & 0xFFFFF000);
	  free_pmem ( table[j] & 0xFFFFF000 );
	}
      }
      free_pmem (table);
    }
  }
}


extern int free_memory;

void task_cleanup(task_t *task)
{
  int i, before, after;
  before = free_memory;
  
  free_page_tables(task->pagedir);
  
  free_pmem(task->pagedir);
  
  free_pmem(task->stack0init);
  free_pmem(task->tss);
  free_pmem(task);

  after = free_memory;

}

unsigned int task_alloc_pmem(int tid, int pages_count)
{
  unsigned int page;
  int i;


  page = alloc_pmem(pages_count);

  if(tid == 0) {
    return page;
  }

  for(i = 0; i < pages_count; i++) {
    add_used_page(tid, page + i*PAGE_SIZE);
  }
  
  return page;
}


void add_used_page (int tid, unsigned int page)
{
  task_t *task;
  task = tasks.task[tid];

  //  printf("mm: marking %X as page used by %i (%i's)\n", page, tid,
  //  	 task->usp_c);


 
}


//add new task to the queue
unsigned int add_task(task_image_t task_image,
		      unsigned int page_directory,
		      int pl)
{
  unsigned int selector, tid;
  stack0_t *stack0;
  task_t *task;
  tss_t *tss;
  int i;
  char *p;

  isem_disable();

  
  
  //  page_directory = hal.pagedir;
  
  //note: both entry and page_directory MUST be aligned
  //on a 4kb boundaries


  tid = alloc_tid();

  
  task = (task_t *) alloc_pmem(1);

  
  tss = (tss_t *) task_alloc_pmem(tid, 1); 

  task->tss = tss;

  
  //  selector = alloc_descriptor();
  //  add_tss_descriptor(selector, (unsigned int) tss, 0x89);

  task->pagedir = page_directory;

  task->image = task_image;

  task->stack0init = task_alloc_pmem(tid, mem_to_pages(STACK0_SIZE));

  
  stack0 = task->stack0init + STACK0_SIZE - sizeof(stack0_t);



  if(page_directory == hal.pagedir) {
    stack0->eip = task_image.start;
    stack0->esp = stack0;
  } else {
    //    stack0->eip = USER_TASK_ENTRY;
    stack0->eip = task_image.entry;
    stack0->esp = USER_STACK_START;
  }   
  

  stack0->cs = USER_CODE_SELECTOR;

  stack0->eflags = 0x202;
  stack0->ss = USER_DATA_SELECTOR;
  stack0->ds = USER_DATA_SELECTOR;
  stack0->es = USER_DATA_SELECTOR;
  stack0->fs = USER_DATA_SELECTOR;
  stack0->gs = USER_DATA_SELECTOR;


  task->stack0 = stack0;

  
  task->signal = 0;
  task->state = TS_READY;
  task->reason_why = RS_NONE;

  task->msg_flag = 0;
 
  task->pl = pl;

  task->msgn = 0;


  task->priority = 1000; //does it matter much?
  
  tasks.task[tid] = task;

  tasks.queue[tasks.count] = tid;
  
  tasks.count++;
  
  
  
  //    printf("Creating Task (ID: %i, QID: %i)... New stack0: %X\n"
  //	   "Entry point: %X, PAGEDIR: %X\n", tid, tasks.count-1, stack0, entry, page_directory);
  //    display_stack0(stack0);
  
  isem_enable();
  
  return tid;
}



void init_taskman()
{
  
  printf("Initializing task manager...\n");
  
  //At first, we initialize system tasks table
  tasks.count = 0;
  tasks.current = 0;
  tasks.current_qid = 0;
  //Memory is allocated for 1024 tasks
  tasks.task = (task_t **) alloc_pmem(1);

  tasks.task[0] = alloc_pmem(1);
  
  tasks.task[0]->pagedir = hal.pagedir;


  init_tids_allocator();
  
  //SYS_TSS - current system process
  
  sys_tss = (tss_t *) alloc_pmem(1);
  sys_tss->esp0 = alloc_pmem(1) + 0x1000 - 4;
  sys_tss->ss0 = SYS_DATA_SELECTOR;
  
  add_tss_descriptor (SYS_TSS, (unsigned int) sys_tss, 0x89);
  

  load_task_register (SYS_TSS);
  
  

  outportb(0x43, 0x34);
  outportb(0x40, 0xfe);
  outportb(0x40, 0x2b);

  
  i_install (0x78, &int_78, RING0 | INTERRUPT_GATE);
}


void load_task_register(int selector)
{
  printf("Loading Task Register with %X\n", selector);
  asm("movl %0, %%eax \n ltr %%ax"::"d"(selector));
}

void display_gdt_descriptor(unsigned int selector)
{
  char *gdt_table;
  int i;

  gdt_table = (char *) hal.gdt_addr;
  
  printf("Descriptor in (%X): ",  selector);
  
  for (i = 0; i < 8; i++) {
    printf("%x", *(gdt_table+selector+i));
  }
  
}

void i_install_task_gate(unsigned char vector, unsigned short task) 
{
  char * idt_table= (char *) IDT_TABLE;
  
  unsigned char i;
  unsigned char b[8];

  b[0]=0;
  b[1]=0;
  b[2]=task & 0x000000FF;
  b[3]=(task & 0x0000FF00) >> 8;
  b[4]=0;
  b[5]=0x85;
  b[6]=0;
  b[7]=0;
  
  
  printf("New Task Gate for vector %i: ", vector);
  for(i=0;i<8;i++){
    *(idt_table+vector*8+i)=b[i];
    printf("%x", b[i]);
  }
  printf("\n");
  
}



char gdt_addr[6];

void set_busy_flag(unsigned int selector)
{
  char * gdt_table;
  char sb;
  
  gdt_table = (char *) hal.gdt_addr;
  
  sb = *(gdt_table + selector + 5);
  
  sb |= 2;
  
  *(gdt_table + selector + 5) = sb;
}

void clear_busy_flag(unsigned int selector)
{
  char * gdt_table;
  char sb;
  
  gdt_table = (char *) hal.gdt_addr;
  
  sb = *(gdt_table + selector + 5);
  
  sb &= 0xfd;
  
  *(gdt_table + selector + 5) = sb;

}

void add_tss_descriptor(unsigned int selector, unsigned int base, unsigned char type)
{
  char * gdt_table;
  unsigned char i;
  unsigned char b[8];

  gdt_table = (char *) hal.gdt_addr;
  
  
  
  b[0]=0x00;
  b[1]=0x10;
  b[2]=( base & 0x000000FF );
  b[3]=( base & 0x0000FF00 ) >> 8;
  b[4]=( base & 0x00FF0000 ) >> 16;
  b[5]=type;
  b[6]=0x40;
  b[7]=( base & 0xFF000000) >> 24;
  
  //  printf("New TSS Descriptor %X (base %X): %x%x%x%x%x%x%x%x\n",
  //	 selector, base, b[0], b[1], b[2], b[3], b[4], b[5], b[6], b[7]);
 
  for(i=0;i<8;i++){
    *(gdt_table+selector+i)=b[i];
  }



}

struct task_ids_table {
  unsigned short tid[0x1000];
  unsigned int pointer;
} tids_table;

void init_tids_allocator(void)
{
  int i;
  
  tids_table.pointer = 0;

  for (i = 1000; i >= 0; i--) {
    free_tid(i);
  }
  
}

void free_tid(unsigned int tid)
{
  tids_table.pointer++;
  tids_table.tid[tids_table.pointer] = tid;
}

unsigned int alloc_tid(void)
{
  unsigned int tid;

  tid = tids_table.tid[tids_table.pointer];
  tids_table.pointer--;
  return tid;
}

int  tick = 0;
char slash = '-';

void c_int_timer()
{
tick++;
if(tick > 18)
 {
 tick = 0;
 switch(slash)
  {
  case '-':
  slash = '\\';
  break;
  case '\\':
  slash = '|';
  break;
  case '|':
  slash = '/';
  break;
  case '/':
  slash = '-';
  break;
  }
 }
char *video = VIDEO_RAM;
*(video + 158) = slash;
}