/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

/*
  vm.c - virtual memory manager (i hope it will be someday ;))
*/

#include <tyros.h>

extern int vend_of_heap;

void init_vm()
{
  set_pte (vend_of_heap, hal.pagedir, 0x7, VM_PAGE_ALLOCATED);
}

int vm_pagefault(unsigned int cr2)
{
  unsigned int pagedir;
  int err;
  unsigned int pte;
  unsigned int ppage;
  unsigned int new_addr;
  unsigned int pde_addr;

  isem_disable();


  pagedir = tasks.task[tasks.current]->pagedir;

  err = get_pde (cr2, pagedir);
  
  
  if( !(err & 1)) { //if pagetable does not exist
    pde_addr = task_alloc_pmem(tasks.current, 1);
    
    set_pde (cr2, pagedir, pde_addr | 0x7);
    
    memsetl(pde_addr, VM_PAGE_ALLOCATED, 1024);
  }

  err = get_pte (cr2, pagedir);
    
  if(err == -1) {
    
    return VM_UNRESOLVED;
  }
  
  pte = err;
  
  if(pte == VM_PAGE_ALLOCATED) {
    //page is allocated virtually, but not allocated physically
    ppage = task_alloc_pmem(tasks.current, 1);


    //    printf("Task %i: Pagefault at %X resolved, allocated %X\n",
    //	   tasks.current, cr2, ppage);
    
    map_memory_page(cr2, ppage, pagedir, 0x7);
    
    
    //    load_cr3(pagedir);
    isem_enable();
    return 0;
  }
    
  if(pte & VM_PAGE_COPY_ON_WRITE){
    //    printf("copy_on_write flag is set\n");
    //now we make physical copy of the page
    new_addr = task_alloc_pmem(tasks.current, 1);
    
    memcpy(new_addr, cr2 & 0xFFFFF000, PAGE_SIZE);
    
    map_memory_page(cr2, new_addr, pagedir, 0x7);
    
    //    printf("Physical copy created, fault resolved\n");
    isem_enable();
    return 0;
  }
  
  isem_enable();
  return VM_UNRESOLVED;
}
