/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

/*
  newbie memory allocator :)
  
  TODO: make it MUCH FASTER
  TODO: make it tread-safe
*/

#include <tyros.h>
#include <ntx.h>
#include <assert.h>

typedef unsigned int size_t;

void display_memory_map(char *buffer);
void *malloc (size_t size);
void *grow(int npages);
void *realloc (void *ptr, size_t size);
void free (void *ptr);

#define LAST_LINK -1 //last link marker


/*
  memory block information structure 
  (situtated in the beginning of each memory block)
*/
typedef struct mb_st {
  size_t size;
  struct mb_st *link;
} mb_t;

static int avail;
static mb_t avlmb;


//that's for debugging
void malloc_test()
{
  int i;
  int *a, *b, *c, *d;
  

  printf("Running memory manager tests....");
  
  a = malloc(0x1000 * sizeof(int));
  
  for(i = 0; i < 0x1000; i++) {
    a[i] = i;
  }
  
  for(i = 0; i < 0x10; i++) {
    b = malloc(i + 0x10);
    
    a = realloc(a, 0x1000 * sizeof(int) + i);
    
    free(b);
    printf(".");
  }
  
  
  
  for(i = 0; i < 0x1000; i++) {
    if(a[i] != i) {
      printf("Test Failed\n");
      for(;;);
    }
  }

  free(a);

  printf("All Tests Have Been Passed\n");

  display_memory_map(0);
  
  //  for(;;);
}

void display_memory_map(char *buffer)
{
  mb_t *p, *q, *t, *cmb;
  size_t k;
  size_t needsize;
  unsigned int begin_block, end_block, begin_block_size, new_block_size, new_block_begin;
  
  q = &avlmb;

  while (TRUE) {
    
    begin_block = q->link;
    
  
    if(begin_block == LAST_LINK) {
    
      break;
      
    }

    t = begin_block;
    
//    buffer += sprintf(buffer, "%i - %i", begin_block, );
    
        printf("(%X): %i - ",t, t->size);

    if( (unsigned int) t < (unsigned int) q) {
      printf("Memory allocation error: %X < %X\n", t, q);
    }
    

    q = t;
  }

  *buffer = 0;
}

void init_mallocator()
{
  mb_t *cmb;
  
  avlmb.size = 0;
  avlmb.link = alloc_pmem(1);
  
  cmb = avlmb.link;
  
  cmb->size = PAGE_SIZE;
  cmb->link = LAST_LINK;
}

void *malloc (size_t size)
{
  void *allocated;
  mb_t *p, *q, *t, *cmb, *tmb;
  size_t k;
  size_t needsize;
  unsigned int begin_block, end_block, begin_block_size, new_block_size, new_block_begin;

  assert(size != 0);
  
  needsize = sizeof(mb_t) + size;
  
  //printf("Requesting %i bytes of physical memory\n", needsize);
  
  q = &avlmb;

  while (TRUE) {
    
    begin_block = q->link;
    
  
    if(begin_block == LAST_LINK) {
    
      q->link = grow( mem_to_pages(size) );
      begin_block = q->link;
    
      
    }

    t = begin_block;

    //   printf("t->size: %i (need %i) \n", t->size, needsize);

    
    if(t->size > needsize + 8) {

      begin_block_size = t->size;
  
      end_block = begin_block + begin_block_size;
      new_block_size = needsize;
      new_block_begin = end_block - new_block_size;
      allocated = new_block_begin + sizeof(mb_t);
    
      p = new_block_begin;
      p->size = new_block_size;
      
      t->size -= new_block_size;

      //      printf("Changed size of %X to %i\n", t, t->size);

      if(t->size == 0) {
	//	printf("Discarding empty block\n");
	q->link = t->link;
      }
    
      break;
    } else {
      //printf("Too small: new q: %X\n", begin_block);
      q = begin_block;
    }
  }


  tmb = (mb_t *) ( (unsigned int) allocated - sizeof(mb_t));
  //  printf("Allocated %i bytes (ptr: %X)\n", tmb->size, allocated);
  //  for(;;);

  return allocated;
}

void *grow(int npages)
{
  mb_t *mb;
  
  mb = alloc_pmem(npages);


  mb->size = PAGE_SIZE * npages;
  mb->link = LAST_LINK;

  
  return mb;
}


int eip;

/*
asm(
    "\n  .globl realloc"
    "\n  realloc:"
    "\n  popl  eip"
    "\n  call _realloc"
    "\n  pushl eip"
    "\n  ret"
);
*/


void *realloc (void *ptr, size_t size)
{
  void *new;
  mb_t *mb;

  mb = ptr - sizeof(mb_t);

  if(mb->size == 0) {
    printf("ATTEMPT TO REALLOC ZERO-LENGTH MEMORY BLOCK. PTR: %X\n", ptr);
    for(;;);
  }

  new = malloc(size);
  memcpy(new, ptr, size);

  free(ptr);

  return new;
}
 
void free (void *ptr)
{
  mb_t *mb, *q, *p;
  void *alink;
 
  mb = ptr - sizeof(mb_t);
 
  if(mb->size == 0) {
    //    printf("ATTEMPT TO FREE ZERO-LENGTH MEMORY BLOCK\n");
    return;
  }

  //insert this block in the list

  //    mb->link = avlmb.link;
  //    avlmb.link = mb;
  
  p = avlmb.link;
    
  do {
    
    if( ((unsigned int) p < (unsigned int) mb ) &&
	(((unsigned int) p->link > (unsigned int) mb) ||
	 p->link == LAST_LINK)){
      
      //   [ A ]   [ B ]   [ C ] 
      
      //B is linked to C by default
      mb->link = p->link;

      //trying to merge B and C
      if(mb->link != LAST_LINK) { //does C exist?
	if( ((unsigned int) mb + mb->size) == mb->link) {
	  //	  printf("upmerging blocks %X->%i and %X->%i", 
	  //		 mb, mb->size, mb->link, mb->link->size);
	  
	  //setting B->link to C->link
	  //and B->size to B-size + C->size
	  mb->size += mb->link->size;
	  mb->link = mb->link->link;
	  
	  //	  printf("%X->size is set to %i\n", mb, mb->size);
	  
	}
      }
      
      //trying to merge A and B
      if( ((unsigned int) p + p->size) == mb) {

	//	display_memory_map(0);

	//	printf("Merging blocks %X->%i and %X->%i, ",
	//	p, p->size, mb, mb->size);



	//setting A->link = B->link
	//and A->size = A->size + B->size
	p->size += mb->size;
	p->link = mb->link;
	//	printf("%X->size is set to %i\n", p, p->size);


	//	display_memory_map(0);
	
	//	for(;;);
	
	break;
	
      }
      
      //A is linked to B if no merging possible
      p->link = mb;
      
      break;

    } else {
      p = p->link;
    }

  } while (TRUE);
  

}

