/*
 *   string.c - klibc memory functions 
 *
 *   Copyright (c) 2003 Lonesome & OSDev Group
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */  

typedef unsigned int size_t;

#include <tyros.h>
#include <ntx.h>

void *memcpy (void *dest, const void *src, size_t n)
{
  int i;
  
  for(i = 0; i < n; i++) {
    *( (char *) dest + i) = *( (char *) src + i);
  }

  return dest;
}

void *memcpyl (int *dest, const int *src, int n)
{
  int i;
  
  for(i = 0; i < n; i++) {
    dest[i] = src[i];
  }

  return dest;
}


void *memset (void *s, int c, size_t n)
{
  int i;
  
  for(i = 0; i < n; i++) {
    *(char *) (s + i) = c;
  }

  return s;
}

void *memsetl(int *s, int c, int n)
{
  int i;
 
  for(i = 0; i < n; i++) {
    s[i] = c;
  }

  return s;
}


void *calloc (size_t nmemb, size_t size)
{
  void *p;
  
  p = malloc(nmemb*size);
  
  memset (p, 0, nmemb*size);
  
  return p;
}

int
memcmp(const void *s1, const void *s2, size_t n)
{
  if (n != 0)
  {
    const unsigned char *p1 = s1, *p2 = s2;

    do {
      if (*p1++ != *p2++)
        return (*--p1 - *--p2);
    } while (--n != 0);
  }
  return 0;
}


unsigned int mem_to_pages(unsigned int mem)
{
  return ((mem/0x1000) + 1);
} __attribute__((always_inline));

