/*
 *   string.c - klibc string functions 
 *
 *   Copyright (c) 2003 Lonesome & OSDev Group
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */  

#include <ntx.h>

char *strcpy(char *dest, const char *src);
int is_digit(char c);

char *strdup (char *s)
{
  char *news;

  news = malloc(strlen(s));
  strcpy(news, s);

  return news;
}



char *strncpy(char *dest, char *src, int n)
{
  char *start;
  start = dest;

  while(*src != 0 && (dest - start) < n) {
    *dest = *src;
    
    dest++;
    src++;
  }

  *dest = 0;

  return dest;
}

char *strcpy(char *dest, const char *src)
{
  return strncpy(dest, src, 0x10000);
}

char *strncat(char *dest, char *src, int n)
{
  char *start;
  char *ret;
  
  ret = dest;

  while(*dest != 0) {
    dest++;
  }

  start = src;

  while(*src != 0 && (src - start) < n) {
    *dest = *src;
    
    dest++;
    src++;
  }

  *dest = 0;
  return ret;
}

char *strcat(char *dest, const char *src)
{
  return strncat (dest, src, 0x10000);
}

int
strncmp(const char *s1, const char *s2, int n)
{

  if (n == 0)
    return 0;
  do {
    if (*s1 != *s2++)
      return *(unsigned const char *)s1 - *(unsigned const char *)--s2;
    if (*s1++ == 0)
      break;
  } while (--n != 0);
  return 0;
}

char *strchr (char *s, int c)
{
  while(*s != c && *s != 0) 
    s++;
  
  if(*s == 0)
    return 0;
  
  return s;
}

int strcmp(const char *s1, const char *s2)
{
  return strncmp(s1, s2, 0x7FFFFFFF);
}


int strlen(char *str)
{
  int i;

  i = 0;

  while(*str != 0) {
    str++;
    i++;
  }
  
  return i;
}

int atoi(const char *s)
{
  char digit[0x10];
  signed int counter=0;
  signed int i;
  int value = 0;
  
  while(is_digit(*s)){
    digit[counter] = *s - '0';
    s++;
    counter++;
  }

  for(i = counter - 1; i >= 0; i--) {
    value += digit[i] * tens_power(((counter-1)-i));
  }

  return value;
}

int tens_power(int n)
{
  int i, value = 1;
  

  for(i = 1; i <= n; i++) {
    value *= 10;
  }

  return value;

}

int is_digit(char c)
{
  if(c >= '0' && c <= '9') {
    return 1;
  } else {
    return 0;
  }
}


char *strrchr(const char *t, int c) {
  register char ch;
  register const char *l=0;

  ch = c;
  for (;;) {
    if (*t == ch) l=t; if (!*t) return (char*)l; ++t;
    if (*t == ch) l=t; if (!*t) return (char*)l; ++t;
    if (*t == ch) l=t; if (!*t) return (char*)l; ++t;
    if (*t == ch) l=t; if (!*t) return (char*)l; ++t;
  }
  return (char*)l;
}


char *
strstr(const char *s, const char *find)
{
  char c, sc;
  long len;

  if ((c = *find++) != 0)
  {
    len = strlen(find);
    do {
      do {
	if ((sc = *s++) == 0)
	  return 0;
      } while (sc != c);
    } while (strncmp(s, find, len) != 0);
    s--;
  }
  return (char*) s;
}


int __errno_location;

char *
strerror(int errnum)
{
  static char ebuf[40];		/* 64-bit number + slop */
  char *cp;
  int v=1000000, lz=0;

  strcpy(ebuf, "Unknown error: ");
  cp = ebuf + 15;
  if (errnum < 0)
  {
    *cp++ = '-';
    errnum = -errnum;
  }
  while (v)
  {
    int d = errnum / v;
    if (d || lz || (v == 1))
    {
      *cp++ = d+'0';
      lz = 1;
    }
    errnum %= v;
    v /= 10;
  }

  return ebuf;
}
