/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

#include <tyros.h>
#include <ntx.h>
#include <gui.h>
#include <syscalls.h>



int usyscall(int eax, int ebx, int ecx, int edx, int esi, int edi)
{
  int retvalue;
  asm("int $0x77":"=a"(retvalue):"a"(eax), "b"(ebx), "c"(ecx), "d"(edx),"S"(esi),"D"(edi));
  return retvalue;
}


void * umorecore (int n)
{
  return (void *) usyscall(SC_MORECORE, n, 0, 0, 0, 0);
}


/* malloc() uses alloc_pmem() instead of morecore() */
void *alloc_pmem (int n)
{
  return (void *) usyscall(SC_MORECORE, n, 0, 0, 0, 0);
}

void __assert_fail (__const char *__assertion, __const char *__file,
                           unsigned int __line, __const char *__function)
{
  printf("ASSERTION FAILED: %s\nFile: %s\nLine: %i\nFunc: %s\n",
  __assertion, __file, __line, __function);
  exit(1);
}


int fstat (int id, ntx_stat_t *stbuf)
{
  return usyscall (SC_STAT, id, stbuf, 0, 0, 0);
}

int waitpid (int tid, int *status, int options)
{
  return usyscall (SC_WAITPID, tid, status, options, 0, 0);
}

int execve (char *filename, char *argv[], char *envp[])
{
  return usyscall (SC_EXECVE, filename, argv, envp, 0, 0);
}


int ntx_launch(char *file, char *argv[], char *envp[])
{
  return usyscall(SC_LAUNCH, file, argv, envp, 0, 0);
}

void _exit(int n)
{
  exit(n);
}

void exit(int n)
{
  return usyscall(SC_EXIT, n, 0, 0, 0, 0);
}

int chdir (char *path)
{
  return usyscall(SC_CHDIR, path, 0, 0, 0, 0);
}

int fork(void)
{
  return usyscall(SC_FORK, 0, 0, 0, 0, 0);
}

int unlink(char *file)
{
  return usyscall(SC_UNLINK, file, 0, 0, 0, 0);
}

int mkdir (char *file, int mode)
{
  return usyscall(SC_MKDIR, file, mode, 0, 0, 0);
}

int mount (char *device, char *dir, char *fs, int flags, char *fsflags)
{
  return usyscall(SC_MOUNT, device, dir, fs, flags, fsflags);
}

int window(int width, int height)
{
  return usyscall(SC_WINDOW, width, height, 0, 0, 0);
}

int win_textout(int id, int x, int y, char *text)
{
  return usyscall(SC_WIN_TEXTOUT, id, x, y, text, 0);
}

int win_mapout (int id, int x, int y, map_t *map)
{
  return usyscall(SC_WIN_MAPOUT, id, x, y, map, 0);
}

int write(int fd, void *buf, int size)
{
  return usyscall(SC_WRITE, fd, buf, size, 0, 0);
}

int read(int fd, void *buf, int size)
{
  return usyscall(SC_READ, fd, buf, size, 0, 0);
}


int open(char *path, int flags)
{
  return usyscall(SC_OPEN, path, flags, 0, 0, 0);
}

//this is guaranteed to work even before libc is initialized
int gputs(char *s)
{
  int console;
  
  console = open("/dev/vc", 0);
  write(console,s, strlen(s));
  close(console);
}

int close(int id)
{
  return usyscall(SC_CLOSE, id, 0, 0, 0, 0);
}

int putchar(char c)
{
char buff[2];
buff[0] = c;
buff[1] = 0;
write(stdout, buff, 1);
}
